"""Alias Singleton"""

import logging
import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import json_to_dict, validate_filter_json


@click.command()
@click.option('--name', type=str, help='Alias name', required=True)
@click.option(
    '--add',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to ADD to alias',
    default=None,
)
@click.option(
    '--remove',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to REMOVE from alias',
    default=None,
)
@click.option(
    '--warn_if_no_indices',
    is_flag=True,
    help='Do not raise exception if there are no actionable indices in add/remove',
)
@click.option(
    '--extra_settings',
    help='JSON version of extra_settings (see documentation)',
    callback=json_to_dict,
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.pass_context
def alias(
    ctx, name, add, remove, warn_if_no_indices, extra_settings, allow_ilm_indices
):
    """
    Add/Remove Indices to/from Alias
    """
    logger = logging.getLogger('cli.singleton.alias')
    manual_options = {
        'name': name,
        'extra_settings': extra_settings,
        'allow_ilm_indices': allow_ilm_indices,
    }
    logger.debug('manual_options %s', manual_options)
    # ctx.info_name is the name of the function or name specified in
    # @click.command decorator
    ignore_empty_list = warn_if_no_indices
    action = CLIAction(
        ctx.info_name,
        ctx.obj['configdict'],
        manual_options,
        [],  # filter_list is empty in our case
        ignore_empty_list,
        add=add,
        remove=remove,
        warn_if_no_indices=warn_if_no_indices,  # alias specific kwargs
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""Main CLI for Curator"""

import sys
import logging
import click
from es_client.defaults import OPTION_DEFAULTS
from es_client.helpers.config import (
    cli_opts,
    context_settings,
    generate_configdict,
    get_client,
    get_config,
    options_from_dict,
)
from es_client.helpers.logging import configure_logging
from es_client.helpers.utils import option_wrapper, prune_nones
from curator.exceptions import ClientException
from curator.classdef import ActionsFile
from curator.defaults.settings import (
    CLICK_DRYRUN,
    VERSION_MAX,
    VERSION_MIN,
    default_config_file,
    footer,
    snapshot_actions,
)
from curator.exceptions import NoIndices, NoSnapshots
from curator.helpers.testers import ilm_policy_check
from curator._version import __version__

ONOFF = {'on': '', 'off': 'no-'}
click_opt_wrap = option_wrapper()

# pylint: disable=R0913, R0914, W0613, W0622, W0718


def ilm_action_skip(client, action_def):
    """
    Skip rollover action if ``allow_ilm_indices`` is ``false``. For all other
    non-snapshot actions, add the ``ilm`` filtertype to the
    :py:attr:`~.curator.ActionDef.filters` list.

    :param action_def: An action object
    :type action_def: :py:class:`~.curator.classdef.ActionDef`

    :returns: ``True`` if ``action_def.action`` is ``rollover`` and the alias
        identified by ``action_def.options['name']`` is associated with an ILM
        policy. This hacky work-around is because the Rollover action does not
        use :py:class:`~.curator.IndexList`
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    if not action_def.allow_ilm and action_def.action not in snapshot_actions():
        if action_def.action == 'rollover':
            if ilm_policy_check(client, action_def.options['name']):
                logger.info(
                    'Alias %s is associated with ILM policy.',
                    action_def.options['name'],
                )
                return True
        elif action_def.filters:
            action_def.filters.append({'filtertype': 'ilm'})
        else:
            action_def.filters = [{'filtertype': 'ilm'}]
    return False


def exception_handler(action_def, err):
    """Do the grunt work with the exception

    :param action_def: An action object
    :param err: The exception

    :type action_def: :py:class:`~.curator.classdef.ActionDef`
    :type err: :py:exc:`Exception`
    """
    logger = logging.getLogger(__name__)
    if isinstance(err, (NoIndices, NoSnapshots)):
        if action_def.iel:
            logger.info(
                'Skipping action "%s" due to empty list: %s',
                action_def.action,
                type(err),
            )
        else:
            logger.error(
                'Unable to complete action "%s".  No actionable items in list: %s',
                action_def.action,
                type(err),
            )
            sys.exit(1)
    else:
        logger.error(
            'Failed to complete action: %s.  %s: %s', action_def.action, type(err), err
        )
        if action_def.cif:
            logger.info(
                'Continuing execution with next action because "continue_if_exception" '
                'is set to True for action %s',
                action_def.action,
            )
        else:
            sys.exit(1)


def process_action(client, action_def, dry_run=False):
    """
    Do the ``action`` in ``action_def.action``, using the associated options and
    any ``kwargs``.

    :param client: A client connection object
    :param action_def: The ``action`` object

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type action_def: :py:class:`~.curator.classdef.ActionDef`
    :rtype: None
    """
    logger = logging.getLogger(__name__)
    logger.debug('Configuration dictionary: %s', action_def.action_dict)
    mykwargs = {}

    logger.debug('INITIAL Action kwargs: %s', mykwargs)
    # Add some settings to mykwargs...
    if action_def.action == 'delete_indices':
        mykwargs['master_timeout'] = 30

    # Update the defaults with whatever came with opts, minus any Nones
    mykwargs.update(prune_nones(action_def.options))

    # Pop out the search_pattern option, if present.
    search_pattern = mykwargs.pop('search_pattern', '*')

    logger.debug('Action kwargs: %s', mykwargs)
    logger.debug('Post search_pattern Action kwargs: %s', mykwargs)

    # Set up the action
    logger.debug('Running "%s"', action_def.action.upper())
    if action_def.action == 'alias':
        # Special behavior for this action, as it has 2 index lists
        action_def.instantiate('action_cls', **mykwargs)
        action_def.instantiate('alias_adds', client, search_pattern=search_pattern)
        action_def.instantiate('alias_removes', client, search_pattern=search_pattern)
        if 'remove' in action_def.action_dict:
            logger.debug('Removing indices from alias "%s"', action_def.options['name'])
            action_def.alias_removes.iterate_filters(action_def.action_dict['remove'])
            action_def.action_cls.remove(
                action_def.alias_removes,
                warn_if_no_indices=action_def.options['warn_if_no_indices'],
            )
        if 'add' in action_def.action_dict:
            logger.debug('Adding indices to alias "%s"', action_def.options['name'])
            action_def.alias_adds.iterate_filters(action_def.action_dict['add'])
            action_def.action_cls.add(
                action_def.alias_adds,
                warn_if_no_indices=action_def.options['warn_if_no_indices'],
            )
    elif action_def.action in ['cluster_routing', 'create_index', 'rollover']:
        action_def.instantiate('action_cls', client, **mykwargs)
    else:
        if action_def.action in ['delete_snapshots', 'restore']:
            mykwargs.pop('repository')  # We don't need to send this value to the action
            action_def.instantiate(
                'list_obj', client, repository=action_def.options['repository']
            )
        else:
            action_def.instantiate('list_obj', client, search_pattern=search_pattern)
        action_def.list_obj.iterate_filters({'filters': action_def.filters})
        logger.debug('Pre Instantiation Action kwargs: %s', mykwargs)
        action_def.instantiate('action_cls', action_def.list_obj, **mykwargs)
    # Do the action
    if dry_run:
        action_def.action_cls.do_dry_run()
    else:
        logger.debug('Doing the action here.')
        action_def.action_cls.do_action()


def run(ctx: click.Context) -> None:
    """
    :param ctx: The Click command context

    :type ctx: :py:class:`Context <click.Context>`

    Called by :py:func:`cli` to execute what was collected at the command-line
    """
    logger = logging.getLogger(__name__)
    logger.debug('action_file: %s', ctx.params['action_file'])
    all_actions = ActionsFile(ctx.params['action_file'])
    for idx in sorted(list(all_actions.actions.keys())):
        action_def = all_actions.actions[idx]
        # Skip to next action if 'disabled'
        if action_def.disabled:
            logger.info(
                'Action ID: %s: "%s" not performed because "disable_action" '
                'is set to True',
                idx,
                action_def.action,
            )
            continue
        logger.info('Preparing Action ID: %s, "%s"', idx, action_def.action)

        # Override the timeout, if specified, otherwise use the default.
        if action_def.timeout_override:
            ctx.obj['configdict']['elasticsearch']['client'][
                'request_timeout'
            ] = action_def.timeout_override

        # Create a client object for each action...
        logger.info('Creating client object and testing connection')

        try:
            client = get_client(
                configdict=ctx.obj['configdict'],
                version_max=VERSION_MAX,
                version_min=VERSION_MIN,
            )
        except ClientException as exc:
            # No matter where logging is set to go, make sure we dump these messages to
            # the CLI
            click.echo('Unable to establish client connection to Elasticsearch!')
            click.echo(f'Exception: {exc}')
            sys.exit(1)
        except Exception as other:
            logger.debug('Fatal exception encountered: %s', other)

        # Filter ILM indices unless expressly permitted
        if ilm_action_skip(client, action_def):
            continue
        #
        # Process the action
        #
        msg = (
            f'Trying Action ID: {idx}, "{action_def.action}": {action_def.description}'
        )
        try:
            logger.info(msg)
            process_action(client, action_def, dry_run=ctx.params['dry_run'])
        except Exception as err:
            exception_handler(action_def, err)
        logger.info('Action ID: %s, "%s" completed.', idx, action_def.action)
    logger.info('All actions completed.')


@click.command(
    context_settings=context_settings(),
    epilog=footer(__version__, tail='command-line.html'),
)
@options_from_dict(OPTION_DEFAULTS)
@click_opt_wrap(*cli_opts('dry-run', settings=CLICK_DRYRUN))
@click.argument('action_file', type=click.Path(exists=True), nargs=1)
@click.version_option(__version__, '-v', '--version', prog_name="curator")
@click.pass_context
def cli(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
    blacklist,
    dry_run,
    action_file,
):
    """
    Curator for Elasticsearch indices

    The default $HOME/.curator/curator.yml configuration file (--config)
    can be used but is not needed.

    Command-line settings will always override YAML configuration settings.

    Some less-frequently used client configuration options are now hidden. To see the
    full list,

    run:

        curator_cli -h
    """
    ctx.obj = {}
    ctx.obj['default_config'] = default_config_file()
    get_config(ctx)
    configure_logging(ctx)
    generate_configdict(ctx)
    run(ctx)



"""Utilities/Helpers for defaults and schemas"""

from os import path
from voluptuous import Any, Boolean, Coerce, Optional
from curator.exceptions import CuratorException

# pylint: disable=E1120

CURATOR_DOCS = 'https://www.elastic.co/guide/en/elasticsearch/client/curator'
CLICK_DRYRUN = {
    'dry-run': {'help': 'Do not perform any changes.', 'is_flag': True},
}
DATA_NODE_ROLES = ['data', 'data_content', 'data_hot', 'data_warm']
EXCLUDE_SYSTEM = (
    '-.kibana*,-.security*,-.watch*,-.triggered_watch*,'
    '-.ml*,-.geoip_databases*,-.logstash*,-.tasks*'
)
VERSION_MIN = (7, 14, 0)
VERSION_MAX = (8, 99, 99)

# Click specifics


def footer(version, tail='index.html'):
    """
    Generate a footer linking to Curator docs based on Curator version

    :param version: The Curator version

    :type version: str

    :returns: An epilog/footer suitable for Click
    """
    if not isinstance(version, str):
        raise CuratorException('Parameter version is not a string: {type(version)}')
    majmin = ''
    try:
        ver = version.split('.')
        majmin = f'{ver[0]}.{ver[1]}'
    except Exception as exc:
        msg = f'Could not determine Curator version from provided value: {version}'
        raise CuratorException(msg) from exc
    return f'Learn more at {CURATOR_DOCS}/{majmin}/{tail}'


# Default Config file location
def default_config_file():
    """
    :returns: The default configuration file location:
        path.join(path.expanduser('~'), '.curator', 'curator.yml')
    """
    default = path.join(path.expanduser('~'), '.curator', 'curator.yml')
    if path.isfile(default):
        return default


# Default filter patterns (regular expressions)
def regex_map():
    """
    :returns: A dictionary of pattern filter 'kind's with their associated regular
        expression: {'timestring': r'^.*{0}.*$', 'regex': r'{0}',
        'prefix': r'^{0}.*$', 'suffix': r'^.*{0}$'}
    """
    return {
        'timestring': r'^.*{0}.*$',
        'regex': r'{0}',
        'prefix': r'^{0}.*$',
        'suffix': r'^.*{0}$',
    }


def date_regex():
    """
    :returns: A dictionary/map of the strftime string characters and their string
        lengths: {'Y':'4', 'G':'4', 'y':'2', 'm':'2', 'W':'2', 'V':'2', 'U':'2',
        'd':'2', 'H':'2', 'M':'2', 'S':'2', 'j':'3'}
    """
    return {
        'Y': '4',
        'G': '4',
        'y': '2',
        'm': '2',
        'W': '2',
        'V': '2',
        'U': '2',
        'd': '2',
        'H': '2',
        'M': '2',
        'S': '2',
        'j': '3',
    }


# Actions


def cluster_actions():
    """
    :returns: A list of supported cluster actions (right now, that's only
        ['cluster_routing'])
    """
    return ['cluster_routing']


def index_actions():
    """
    :returns: The list of supported index actions:
        [ 'alias', 'allocation', 'close', 'create_index', 'delete_indices',
        'forcemerge', 'index_settings', 'open', 'reindex', 'replicas',
        'rollover', 'shrink', 'snapshot']
    """
    return [
        'alias',
        'allocation',
        'close',
        'cold2frozen',
        'create_index',
        'delete_indices',
        'forcemerge',
        'index_settings',
        'open',
        'reindex',
        'replicas',
        'rollover',
        'shrink',
        'snapshot',
    ]


def snapshot_actions():
    """
    :returns: The list of supported snapshot actions: ['delete_snapshots', 'restore']
    """
    return ['delete_snapshots', 'restore']


def all_actions():
    """
    :returns: A sorted list of all supported actions: cluster, index, and snapshot
    """
    return sorted(cluster_actions() + index_actions() + snapshot_actions())


def index_filtertypes():
    """
    :returns: The list of supported index filter types:
        ['alias', 'allocated', 'age', 'closed', 'count', 'empty', 'forcemerged',
        'ilm', 'kibana', 'none', 'opened', 'pattern', 'period', 'space',
        'shards', 'size']
    """

    return [
        'alias',
        'allocated',
        'age',
        'closed',
        'count',
        'empty',
        'forcemerged',
        'ilm',
        'kibana',
        'none',
        'opened',
        'pattern',
        'period',
        'space',
        'shards',
        'size',
    ]


def snapshot_filtertypes():
    """
    :returns: The list of supported snapshot filter types: ['age', 'count', 'none',
        'pattern', 'period', 'state']
    """
    return ['age', 'count', 'none', 'pattern', 'period', 'state']


def all_filtertypes():
    """
    :returns: A sorted list of all supported filter types (both snapshot and index)
    """
    return sorted(list(set(index_filtertypes() + snapshot_filtertypes())))


def default_options():
    """
    :returns: The default values for these options:
        {'allow_ilm_indices': False, 'continue_if_exception': False,
        'disable_action': False, 'ignore_empty_list': False,
        'timeout_override': None}
    """
    return {
        'allow_ilm_indices': False,
        'continue_if_exception': False,
        'disable_action': False,
        'ignore_empty_list': False,
        'timeout_override': None,
    }


def default_filters():
    """
    If no filters are set, add a 'none' filter

    :returns: {'filters': [{'filtertype': 'none'}]}
    """
    return {'filters': [{'filtertype': 'none'}]}


def structural_filter_elements():
    """
    :returns: Barebones schemas for initial validation of filters
    """

    return {
        Optional('aliases'): Any(list, str),
        Optional('allocation_type'): Any(str),
        Optional('count'): Coerce(int),
        Optional('date_from'): Any(None, str),
        Optional('date_from_format'): Any(None, str),
        Optional('date_to'): Any(None, str),
        Optional('date_to_format'): Any(None, str),
        Optional('direction'): Any(str),
        Optional('disk_space'): float,
        Optional('epoch'): Any(Coerce(int), None),
        Optional('exclude'): Any(None, bool, int, str),
        Optional('field'): Any(None, str),
        Optional('intersect'): Any(None, bool, int, str),
        Optional('key'): Any(str),
        Optional('kind'): Any(str),
        Optional('max_num_segments'): Coerce(int),
        Optional('number_of_shards'): Coerce(int),
        Optional('pattern'): Any(str),
        Optional('period_type'): Any(str),
        Optional('reverse'): Any(None, bool, int, str),
        Optional('range_from'): Coerce(int),
        Optional('range_to'): Coerce(int),
        Optional('shard_filter_behavior'): Any(str),
        Optional('size_behavior'): Any(str),
        Optional('size_threshold'): Any(Coerce(float)),
        Optional('source'): Any(str),
        Optional('state'): Any(str),
        Optional('stats_result'): Any(None, str),
        Optional('timestring'): Any(None, str),
        Optional('threshold_behavior'): Any(str),
        Optional('unit'): Any(str),
        Optional('unit_count'): Coerce(int),
        Optional('unit_count_pattern'): Any(str),
        Optional('use_age'): Boolean(),
        Optional('value'): Any(int, float, bool, str),
        Optional('week_starts_on'): Any(None, str),
    }



"""Snapshot Singleton"""

import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import validate_filter_json


# pylint: disable=line-too-long
@click.command()
@click.option(
    '--search_pattern', type=str, default='*', help='Elasticsearch Index Search Pattern'
)
@click.option('--repository', type=str, required=True, help='Snapshot repository')
@click.option(
    '--name',
    type=str,
    help='Snapshot name',
    show_default=True,
    default='curator-%Y%m%d%H%M%S',
)
@click.option(
    '--ignore_unavailable',
    is_flag=True,
    show_default=True,
    help='Ignore unavailable shards/indices.',
)
@click.option(
    '--include_global_state',
    is_flag=True,
    show_default=True,
    help='Store cluster global state with snapshot.',
)
@click.option(
    '--partial',
    is_flag=True,
    show_default=True,
    help='Do not fail if primary shard is unavailable.',
)
@click.option(
    '--wait_for_completion/--no-wait_for_completion',
    default=True,
    show_default=True,
    help='Wait for the snapshot to complete',
)
@click.option(
    '--wait_interval',
    default=9,
    type=int,
    help='Seconds to wait between completion checks.',
)
@click.option(
    '--max_wait',
    default=-1,
    type=int,
    help='Maximum number of seconds to wait_for_completion',
)
@click.option(
    '--skip_repo_fs_check',
    is_flag=True,
    show_default=True,
    help='Skip repository filesystem access validation.',
)
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable indices',
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to act on.',
    required=True,
)
@click.pass_context
def snapshot(
    ctx,
    search_pattern,
    repository,
    name,
    ignore_unavailable,
    include_global_state,
    partial,
    skip_repo_fs_check,
    wait_for_completion,
    wait_interval,
    max_wait,
    ignore_empty_list,
    allow_ilm_indices,
    filter_list,
):
    """
    Snapshot Indices
    """
    manual_options = {
        'search_pattern': search_pattern,
        'name': name,
        'repository': repository,
        'ignore_unavailable': ignore_unavailable,
        'include_global_state': include_global_state,
        'partial': partial,
        'skip_repo_fs_check': skip_repo_fs_check,
        'wait_for_completion': wait_for_completion,
        'max_wait': max_wait,
        'wait_interval': wait_interval,
        'allow_ilm_indices': allow_ilm_indices,
    }
    # ctx.info_name is the name of the function or name specified in
    # @click.command decorator
    action = CLIAction(
        ctx.info_name,
        ctx.obj['configdict'],
        manual_options,
        filter_list,
        ignore_empty_list,
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""Functions validating the ``filter`` Schema of an ``action``"""

import logging
from voluptuous import Any, In, Required, Schema
from es_client.helpers.schemacheck import SchemaCheck
from es_client.helpers.utils import prune_nones
from curator.defaults import settings, filtertypes
from curator.exceptions import ConfigurationError

logger = logging.getLogger(__name__)


def filtertype():
    """
    Return a :py:class:`~.voluptuous.schema_builder.Schema` object that uses
    :py:func:`~.curator.defaults.settings.all_filtertypes` to populate acceptable values

    :returns: A :py:class:`~.voluptuous.schema_builder.Schema` object
    """
    return {
        Required('filtertype'): Any(
            In(settings.all_filtertypes()),
            msg=f'filtertype must be one of {settings.all_filtertypes()}',
        )
    }


def filterstructure():
    """
    Return a :py:class:`~.voluptuous.schema_builder.Schema` object that uses the
    return value from :py:func:`~.curator.defaults.settings.structural_filter_elements`
    to populate acceptable values and updates/merges the Schema object with the
    return value from :py:func:`filtertype`

    :returns: A :py:class:`~.voluptuous.schema_builder.Schema` object
    """
    # This is to first ensure that only the possible keys/filter elements are
    # there, and get a dictionary back to work with.
    retval = settings.structural_filter_elements()
    retval.update(filtertype())
    return Schema(retval)


def singlefilter(action, data):
    """
    Return a :py:class:`~.voluptuous.schema_builder.Schema` object that is created
    using the return value from :py:func:`filtertype` to create a local variable
    ``ftype``. The values from ``action`` and ``data`` are used to update ``ftype``
    based on matching function names in :py:mod:`~.curator.defaults.filtertypes`.

    :py:func:`~.curator.defaults.settings.structural_filter_elements` to populate
    acceptable values and updates/merges the Schema object with the return value
    from :py:func:`filtertype`

    :param action: The Curator action name
    :type action: str
    :param data: The filter block of the action

    :returns: A :py:class:`~.voluptuous.schema_builder.Schema` object
    """
    try:
        ftdata = data['filtertype']
    except KeyError as exc:
        raise ConfigurationError('Missing key "filtertype"') from exc
    ftype = filtertype()
    for each in getattr(filtertypes, ftdata)(action, data):
        ftype.update(each)
    return Schema(ftype)


def validfilters(action, location=None):
    """Validate the filters in a list"""

    def func(val):
        """This validator method simply validates all filters in the list."""
        for idx, value in enumerate(val):
            pruned = prune_nones(value)
            filter_dict = SchemaCheck(
                pruned,
                singlefilter(action, pruned),
                'filter',
                f'{location}, filter #{idx}: {pruned}',
            ).result()
            logger.debug('Filter #%s: %s', idx, filter_dict)
            val[idx] = filter_dict
        # If we've made it here without raising an Exception, it's valid
        return val

    return func



"""Filtertype schema definitions"""

import logging
from curator.defaults import filter_elements, settings

# pylint: disable=missing-docstring, unused-argument, line-too-long


def _age_elements(action, config):
    """
    Sort which filter types that have ``use_age`` are suitable for
    :py:class:`~.curator.IndexList` and which are acceptable in
    :py:class:`~.curator.SnapshotList`, which are required, and which are not.

    :param action: The name of an action
    :type action: str
    :param config: The configuration block for one filter of ``action``
    :type config: dict

    :returns: A :py:class:`list` containing one or more
        :py:class:`~.voluptuous.schema_builder.Optional` or
        :py:class:`~.voluptuous.schema_builder.Required` options from
        :py:mod:`~.curator.defaults.filter_elements`, defining acceptable values
        for each for the given ``action``
    :rtype: list
    """
    retval = []
    is_req = True
    if config['filtertype'] in ['count', 'space']:
        # is_req = True if 'use_age' in config and config['use_age'] else False
        is_req = bool('use_age' in config and config['use_age'])
    retval.append(filter_elements.source(action=action, required=is_req))
    if action in settings.index_actions():
        retval.append(filter_elements.stats_result())
    # This is a silly thing here, because the absence of 'source' will
    # show up in the actual schema check, but it keeps code from breaking here
    ts_req = False
    if 'source' in config:
        if config['source'] == 'name':
            ts_req = True
        elif action in settings.index_actions():
            # field_stats must _only_ exist for Index actions (not Snapshot)
            if config['source'] == 'field_stats':
                retval.append(filter_elements.field(required=True))
            else:
                retval.append(filter_elements.field(required=False))
        retval.append(filter_elements.timestring(required=ts_req))
    else:
        # If source isn't in the config, then the other elements are not
        # required, but should be Optional to prevent false positives
        retval.append(filter_elements.field(required=False))
        retval.append(filter_elements.timestring(required=ts_req))
    return retval


# ### Schema information ###


def alias(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_by_alias`
    """
    return [
        filter_elements.aliases(),
        filter_elements.exclude(),
    ]


def age(action, config):
    """
    :returns: Filter elements acceptable for
        :py:class:`~.curator.IndexList` :py:meth:`~.curator.IndexList.filter_by_age` or
        :py:class:`~.curator.SnapshotList`
        :py:meth:`~.curator.SnapshotList.filter_by_age`
    """
    # Required & Optional
    logger = logging.getLogger('curator.defaults.filtertypes.age')
    retval = [
        filter_elements.direction(),
        filter_elements.unit(),
        filter_elements.unit_count(),
        filter_elements.unit_count_pattern(),
        filter_elements.epoch(),
        filter_elements.exclude(),
    ]
    retval += _age_elements(action, config)
    logger.debug('AGE FILTER = %s', retval)
    return retval


def allocated(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_allocated`
    """
    return [
        filter_elements.key(),
        filter_elements.value(),
        filter_elements.allocation_type(),
        filter_elements.exclude(exclude=True),
    ]


def closed(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_closed`
    """
    return [filter_elements.exclude(exclude=True)]


def count(action, config):
    """
    :returns: Filter elements acceptable for
        :py:class:`~.curator.IndexList` :py:meth:`~.curator.IndexList.filter_by_count`
        or :py:class:`~.curator.SnapshotList`
        :py:meth:`~.curator.SnapshotList.filter_by_count`
    """
    retval = [
        filter_elements.count(),
        filter_elements.use_age(),
        filter_elements.pattern(),
        filter_elements.reverse(),
        filter_elements.exclude(exclude=True),
    ]
    retval += _age_elements(action, config)
    return retval


def empty(action, config):
    """
    :returns: Filter elements acceptable for :py:meth:`~.curator.IndexList.filter_empty`
    """
    return [filter_elements.exclude()]


def forcemerged(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_forceMerged`
    """
    return [
        filter_elements.max_num_segments(),
        filter_elements.exclude(exclude=True),
    ]


def ilm(action, config):
    """
    :returns: Filter elements acceptable for :py:meth:`~.curator.IndexList.filter_ilm`
    """
    return [filter_elements.exclude(exclude=True)]


def kibana(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_kibana`
    """
    return [filter_elements.exclude(exclude=True)]


def none(action, config):
    """
    :returns: Filter elements acceptable for
        :py:class:`~.curator.IndexList` :py:meth:`~.curator.IndexList.filter_none` or
        :py:class:`~.curator.SnapshotList` :py:meth:`~.curator.SnapshotList.filter_none`
    """
    return []


def opened(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_opened`
    """
    return [filter_elements.exclude(exclude=True)]


def pattern(action, config):
    """
    :returns: Filter elements acceptable for
        :py:class:`~.curator.IndexList` :py:meth:`~.curator.IndexList.filter_by_regex`
        or :py:class:`~.curator.SnapshotList`
        :py:meth:`~.curator.SnapshotList.filter_by_regex`
    """
    return [
        filter_elements.kind(),
        filter_elements.value(),
        filter_elements.exclude(),
    ]


def period(action, config):
    """
    :returns: Filter elements acceptable for
        :py:class:`~.curator.IndexList` :py:meth:`~.curator.IndexList.filter_period` or
        :py:class:`~.curator.SnapshotList`
        :py:meth:`~.curator.SnapshotList.filter_period`
    """
    retval = [
        filter_elements.unit(period=True),
        filter_elements.range_from(),
        filter_elements.range_to(),
        filter_elements.week_starts_on(),
        filter_elements.epoch(),
        filter_elements.exclude(),
        filter_elements.period_type(),
        filter_elements.date_from(),
        filter_elements.date_from_format(),
        filter_elements.date_to(),
        filter_elements.date_to_format(),
    ]
    # Only add intersect() to index actions.
    if action in settings.index_actions():
        retval.append(filter_elements.intersect())
    retval += _age_elements(action, config)
    return retval


def shards(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_by_shards`
    """
    return [
        filter_elements.number_of_shards(),
        filter_elements.shard_filter_behavior(),
        filter_elements.exclude(),
    ]


def size(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_by_size`
    """
    return [
        filter_elements.size_threshold(),
        filter_elements.threshold_behavior(),
        filter_elements.size_behavior(),
        filter_elements.exclude(),
    ]


def space(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.IndexList.filter_by_space`
    """
    retval = [
        filter_elements.disk_space(),
        filter_elements.reverse(),
        filter_elements.use_age(),
        filter_elements.exclude(),
        filter_elements.threshold_behavior(),
    ]
    retval += _age_elements(action, config)
    return retval


def state(action, config):
    """
    :returns: Filter elements acceptable for
        :py:meth:`~.curator.SnapshotList.filter_by_state`
    """
    return [
        filter_elements.state(),
        filter_elements.exclude(),
    ]



"""Delete index action class"""

import logging

# pylint: disable=import-error
from curator.helpers.getters import get_indices
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure, show_dry_run, to_csv


class DeleteIndices:
    """Delete Indices Action Class"""

    def __init__(self, ilo, master_timeout=30):
        """
        :param ilo: An IndexList Object
        :param master_timeout: Number of seconds to wait for master node response

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type master_timeout: int
        """
        verify_index_list(ilo)
        if not isinstance(master_timeout, int):
            raise TypeError(
                f'Incorrect type for "master_timeout": {type(master_timeout)}. '
                f'Should be integer value.'
            )
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from
        #: param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: String value of param ``master_timeout`` + ``s``, for seconds.
        self.master_timeout = str(master_timeout) + 's'
        self.loggit = logging.getLogger('curator.actions.delete_indices')
        self.loggit.debug('master_timeout value: %s', self.master_timeout)

    def _verify_result(self, result, count):
        """
        Breakout method to aid readability

        :param result: A list of indices from :py:meth:`__chunk_loop`
        :param count: The number of tries that have occurred

        :type result: list
        :type count: int

        :returns: ``True`` if result is verified successful, else ``False``
        :rtype: bool
        """
        if isinstance(result, list) and result:
            self.loggit.error(
                'The following indices failed to delete on try #%s:', count
            )
            for idx in result:
                self.loggit.error("---%s", idx)
            retval = False
        else:
            self.loggit.debug('Successfully deleted all indices on try #%s', count)
            retval = True
        return retval

    def __chunk_loop(self, chunk_list):
        """
        Loop through deletes 3 times to ensure they complete

        :param chunk_list: A list of indices pre-chunked so it won't overload
            the URL size limit.
        :type chunk_list: list
        """
        working_list = chunk_list
        for count in range(1, 4):  # Try 3 times
            for i in working_list:
                self.loggit.info("---deleting index %s", i)
            self.client.indices.delete(
                index=to_csv(working_list), master_timeout=self.master_timeout
            )
            result = [i for i in working_list if i in get_indices(self.client)]
            if self._verify_result(result, count):
                return
            working_list = result
        self.loggit.error(
            'Unable to delete the following indices after 3 attempts: %s', result
        )

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        show_dry_run(self.index_list, 'delete_indices')

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.delete` indices in
        :py:attr:`index_list`
        """
        self.index_list.empty_list_check()
        msg = (
            f'Deleting {len(self.index_list.indices)} selected indices: '
            f'{self.index_list.indices}'
        )
        self.loggit.info(msg)
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                self.__chunk_loop(lst)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Cluster Routing action class"""

import logging

# pylint: disable=import-error
from curator.helpers.testers import verify_client_object
from curator.helpers.utils import report_failure
from curator.helpers.waiters import wait_for_it


class ClusterRouting:
    """ClusterRouting Action Class"""

    def __init__(
        self,
        client,
        routing_type=None,
        setting=None,
        value=None,
        wait_for_completion=False,
        wait_interval=9,
        max_wait=-1,
    ):
        """
        For now, the cluster routing settings are hardcoded to be ``transient``

        :param client: A client connection object
        :param routing_type: Type of routing to apply. Either ``allocation`` or
            ``rebalance``
        :param setting: Currently, the only acceptable value for ``setting`` is
            ``enable``. This is here in case that changes.
        :param value: Used only if ``setting`` is ``enable``. Semi-dependent on
            ``routing_type``. Acceptable values for ``allocation`` and ``rebalance``
            are ``all``, ``primaries``, and ``none`` (string, not :py:class:`None`).
            If ``routing_type`` is ``allocation``, this can also be ``new_primaries``,
            and if ``rebalance``, it can be ``replicas``.
        :param wait_for_completion: Wait for completion before returning.
        :param wait_interval: Seconds to wait between completion checks.
        :param max_wait: Maximum number of seconds to ``wait_for_completion``

        :type client: :py:class:`~.elasticsearch.Elasticsearch`
        :type routing_type: str
        :type setting: str
        :type value: str
        :type wait_for_completion: bool
        :type wait_interval: int
        :type max_wait: int
        """
        verify_client_object(client)
        #: An :py:class:`~.elasticsearch.Elasticsearch` client object
        self.client = client
        self.loggit = logging.getLogger('curator.actions.cluster_routing')
        #: Object attribute that gets the value of param ``wait_for_completion``
        self.wfc = wait_for_completion
        #: Object attribute that gets the value of param ``wait_interval``
        self.wait_interval = wait_interval
        #: Object attribute that gets the value of param ``max_wait``. How long
        #: in seconds to :py:attr:`wfc` before returning with an exception. A value
        #: of ``-1`` means wait forever.
        self.max_wait = max_wait

        if setting != 'enable':
            raise ValueError(f'Invalid value for "setting": {setting}.')
        if routing_type == 'allocation':
            if value not in ['all', 'primaries', 'new_primaries', 'none']:
                raise ValueError(
                    f'Invalid "value": {value} with "routing_type":{routing_type}.'
                )
        elif routing_type == 'rebalance':
            if value not in ['all', 'primaries', 'replicas', 'none']:
                raise ValueError(
                    f'Invalid "value": {value} with "routing_type": {routing_type}.'
                )
        else:
            raise ValueError(f'Invalid value for "routing_type":{routing_type}.')
        bkey = f'cluster.routing.{routing_type}.{setting}'
        #: Populated at instance creation time. Value is built from the passed
        #: values from params ``routing_type`` and ``setting``, e.g.
        #: ``cluster.routing.routing_type.setting``
        self.settings = {bkey: value}

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        msg = f'DRY-RUN: Update cluster routing transient settings: {self.settings}'
        self.loggit.info(msg)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.ClusterClient.put_settings` to the cluster with
        :py:attr:`settings`.
        """
        self.loggit.info('Updating cluster settings: %s', self.settings)
        try:
            self.client.cluster.put_settings(transient=self.settings)
            if self.wfc:
                self.loggit.debug(
                    'Waiting for shards to complete routing and/or rebalancing'
                )
                wait_for_it(
                    self.client,
                    'cluster_routing',
                    wait_interval=self.wait_interval,
                    max_wait=self.max_wait,
                )
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Alias action"""

import logging

# pylint: disable=import-error
from curator.exceptions import ActionError, MissingArgument, NoIndices
from curator.helpers.date_ops import parse_date_pattern, parse_datemath
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import report_failure


class Alias:
    """Alias Action Class"""

    # pylint: disable=unused-argument
    def __init__(self, name=None, extra_settings=None, **kwargs):
        """
        :param name: The alias name
        :param extra_settings: Extra settings, including filters and routing.
            For more information see `here
            </https://www.elastic.co/guide/en/elasticsearch/reference/8.6/indices-aliases.html>`_.

        :type name: str
        :type extra_settings: dict
        """
        if extra_settings is None:
            extra_settings = {}
        if not name:
            raise MissingArgument('No value for "name" provided.')
        #: The :py:func:`~.curator.helpers.date_ops.parse_date_pattern` rendered
        #: version of what was passed by param ``name``.
        self.name = parse_date_pattern(name)
        #: The list of actions to perform.  Populated by
        #: :py:meth:`~.curator.actions.Alias.add` and
        #: :py:meth:`~.curator.actions.Alias.remove`
        self.actions = []
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object which will
        #: later be set by :py:meth:`~.curator.actions.Alias.add` or
        #: :py:meth:`~.curator.actions.Alias.remove`
        self.client = None
        #: Any extra things to add to the alias, like filters, or routing. Gets
        #: the value from param ``extra_settings``.
        self.extra_settings = extra_settings
        self.loggit = logging.getLogger('curator.actions.alias')
        #: Preset default value to ``False``.
        self.warn_if_no_indices = False

    def add(self, ilo, warn_if_no_indices=False):
        """
        Create ``add`` statements for each index in ``ilo`` for :py:attr:`name`, then
        append them to :py:attr:`actions`.  Add any :py:attr:`extra_settings` that
        may be there.

        :param ilo: An IndexList Object
        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        """
        verify_index_list(ilo)
        self.loggit.debug('ADD -> ILO = %s', ilo.indices)
        if not self.client:
            self.client = ilo.client
        self.name = parse_datemath(self.client, self.name)
        try:
            ilo.empty_list_check()
        except NoIndices as exc:
            # Add a warning if there are no indices to add, if so set in options
            if warn_if_no_indices:
                self.warn_if_no_indices = True
                self.loggit.warning(
                    'No indices found after processing filters. Nothing to add to %s',
                    self.name,
                )
                return
            # Re-raise the exceptions.NoIndices so it will behave as before
            raise NoIndices('No indices to add to alias') from exc
        for index in ilo.working_list():
            self.loggit.debug(
                'Adding index %s to alias %s with extra settings %s',
                index,
                self.name,
                self.extra_settings,
            )
            add_dict = {'add': {'index': index, 'alias': self.name}}
            add_dict['add'].update(self.extra_settings)
            self.actions.append(add_dict)

    def remove(self, ilo, warn_if_no_indices=False):
        """
        Create ``remove`` statements for each index in ``ilo`` for :py:attr:`name`,
        then append them to :py:attr:`actions`.

        :param ilo: An IndexList Object
        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        """
        verify_index_list(ilo)
        self.loggit.debug('REMOVE -> ILO = %s', ilo.indices)
        if not self.client:
            self.client = ilo.client
        self.name = parse_datemath(self.client, self.name)
        try:
            ilo.empty_list_check()
        except NoIndices as exc:
            # Add a warning if there are no indices to add, if so set in options
            if warn_if_no_indices:
                self.warn_if_no_indices = True
                self.loggit.warning(
                    'No indices found after processing filters. '
                    'Nothing to remove from %s',
                    self.name,
                )
                return

            # Re-raise the exceptions.NoIndices so it will behave as before
            raise NoIndices('No indices to remove from alias') from exc
        aliases = self.client.indices.get_alias(expand_wildcards=['open', 'closed'])
        for index in ilo.working_list():
            if index in aliases:
                self.loggit.debug('Index %s in get_aliases output', index)
                # Only remove if the index is associated with the alias
                if self.name in aliases[index]['aliases']:
                    self.loggit.debug(
                        'Removing index %s from alias %s', index, self.name
                    )
                    self.actions.append(
                        {'remove': {'index': index, 'alias': self.name}}
                    )
                else:
                    self.loggit.debug(
                        'Can not remove: Index %s is not associated with alias %s',
                        index,
                        self.name,
                    )

    def check_actions(self):
        """
        :returns: :py:attr:`actions` for use with the
            :py:meth:`~.elasticsearch.client.IndicesClient.update_aliases` API
            call if actions exist, otherwise an exception is raised.
        """
        if not self.actions:
            if not self.warn_if_no_indices:
                raise ActionError('No "add" or "remove" operations')
            raise NoIndices('No "adds" or "removes" found.  Taking no action')
        self.loggit.debug('Alias actions: %s', self.actions)

        return self.actions

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        for item in self.check_actions():
            job = list(item.keys())[0]
            index = item[job]['index']
            alias = item[job]['alias']
            # We want our log to look clever, so if job is "remove", strip the
            # 'e' so "remove" can become "removing".  "adding" works already.
            msg = (
                f"DRY-RUN: alias: {job.rstrip('e')}ing index \"{index}\" "
                f"{'to' if job == 'add' else 'from'} alias \"{alias}\""
            )
            self.loggit.info(msg)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.update_aliases` for
        :py:attr:`name` with :py:attr:`actions`
        """
        self.loggit.info('Updating aliases...')
        self.loggit.info('Alias actions: %s', self.actions)
        try:
            self.client.indices.update_aliases(actions=self.actions)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Index List Class"""

import re
import itertools
import logging
from elasticsearch8.exceptions import NotFoundError, TransportError
from es_client.helpers.schemacheck import SchemaCheck
from es_client.helpers.utils import ensure_list
from curator.defaults import settings
from curator.exceptions import (
    ActionError,
    ConfigurationError,
    MissingArgument,
    NoIndices,
)
from curator.helpers.date_ops import (
    absolute_date_range,
    date_range,
    fix_epoch,
    get_date_regex,
    get_point_of_reference,
    get_unit_count_from_name,
    TimestringSearch,
)
from curator.helpers.getters import byte_size, get_indices
from curator.helpers.testers import verify_client_object
from curator.helpers.utils import chunk_index_list, report_failure, to_csv
from curator.validators.filter_functions import filterstructure


class IndexList:
    """IndexList class"""

    def __init__(self, client, search_pattern='*'):
        verify_client_object(client)
        self.loggit = logging.getLogger('curator.indexlist')
        #: An :py:class:`~.elasticsearch.Elasticsearch` client object passed from
        #: param ``client``
        self.client = client
        #: Information extracted from indices, such as segment count, age, etc.
        #: Populated at instance creation time by private helper methods.
        #: **Type:** :py:class:`dict`
        self.index_info = {}
        #: The running list of indices which will be used by one of the
        #: :py:mod:`~.curator.actions` classes. Populated at instance creation
        #: time by private helper methods. **Type:** :py:class:`list`
        self.indices = []
        #: All indices in the cluster at instance creation time.
        #: **Type:** :py:class:`list`
        self.all_indices = []
        self.__get_indices(search_pattern)
        self.age_keyfield = None

    def __actionable(self, idx):
        self.loggit.debug('Index %s is actionable and remains in the list.', idx)

    def __not_actionable(self, idx):
        self.loggit.debug('Index %s is not actionable, removing from list.', idx)
        self.indices.remove(idx)

    def __excludify(self, condition, exclude, index, msg=None):
        if condition is True:
            if exclude:
                text = "Removed from actionable list"
                self.__not_actionable(index)
            else:
                text = "Remains in actionable list"
                self.__actionable(index)
        else:
            if exclude:
                text = "Remains in actionable list"
                self.__actionable(index)
            else:
                text = "Removed from actionable list"
                self.__not_actionable(index)
        if msg:
            self.loggit.debug('%s: %s', text, msg)

    def __get_indices(self, pattern):
        """
        Pull all indices into ``all_indices``, then populate ``indices`` and
        ``index_info``
        """
        self.loggit.debug('Getting indices matching search_pattern: "%s"', pattern)
        self.all_indices = get_indices(self.client, search_pattern=pattern)
        self.indices = self.all_indices[:]
        # if self.indices:
        #     for index in self.indices:
        #         self.__build_index_info(index)

    def __build_index_info(self, index):
        """
        Ensure that ``index`` is a key in ``index_info``. If not, create a
        sub-dictionary structure under that key.
        """
        self.loggit.debug('Building preliminary index metadata for %s', index)
        if index not in self.index_info:
            self.index_info[index] = self.__zero_values()

    def __map_method(self, ftype):
        methods = {
            'alias': self.filter_by_alias,
            'age': self.filter_by_age,
            'allocated': self.filter_allocated,
            'closed': self.filter_closed,
            'count': self.filter_by_count,
            'empty': self.filter_empty,
            'forcemerged': self.filter_forceMerged,
            'ilm': self.filter_ilm,
            'kibana': self.filter_kibana,
            'none': self.filter_none,
            'opened': self.filter_opened,
            'period': self.filter_period,
            'pattern': self.filter_by_regex,
            'space': self.filter_by_space,
            'shards': self.filter_by_shards,
            'size': self.filter_by_size,
        }
        return methods[ftype]

    def __remove_missing(self, err):
        """
        Remove missing index found in ``err`` from self.indices and return that name
        """
        missing = err.info['error']['index']
        self.loggit.warning('Index was initiallly present, but now is not: %s', missing)
        self.loggit.debug('Removing %s from active IndexList', missing)
        self.indices.remove(missing)
        return missing

    def __zero_values(self):
        """The default values for index metadata"""
        return {
            'age': {'creation_date': 0, 'name': 0},
            'docs': 0,
            'number_of_replicas': 0,
            'number_of_shards': 0,
            'primary_size_in_bytes': 0,
            'routing': {},
            'segments': 0,
            'size_in_bytes': 0,
            'state': '',
        }

    def _get_indices_segments(self, data):
        return self.client.indices.segments(index=to_csv(data))['indices'].copy()

    def _get_indices_settings(self, data):
        return self.client.indices.get_settings(index=to_csv(data))

    def _get_indices_stats(self, data):
        return self.client.indices.stats(index=to_csv(data), metric='store,docs')[
            'indices'
        ]

    def _bulk_queries(self, data, exec_func):
        slice_number = 10
        query_result = {}
        loop_number = (
            round(len(data) / slice_number)
            if round(len(data) / slice_number) > 0
            else 1
        )
        self.loggit.debug("Bulk Queries - number requests created: %s", loop_number)
        for num in range(0, loop_number):
            if num == (loop_number - 1):
                data_sliced = data[num * slice_number :]
            else:
                data_sliced = data[num * slice_number : (num + 1) * slice_number]
            query_result.update(exec_func(data_sliced))
        return query_result

    def mitigate_alias(self, index):
        """
        Mitigate when an alias is detected instead of an index name

        :param index: The index name that is showing up *instead* of what was expected

        :type index: str

        :returns: No return value:
        :rtype: None
        """
        self.loggit.debug('BEGIN mitigate_alias')
        self.loggit.debug(
            'Correcting an instance where an alias name points to index "%s"', index
        )
        data = self.client.indices.get(index=index)
        aliases = list(data[index]['aliases'])
        if aliases:
            for alias in aliases:
                if alias in self.indices:
                    self.loggit.warning(
                        'Removing alias "%s" from IndexList.indices', alias
                    )
                    self.indices.remove(alias)
                if alias in list(self.index_info):
                    self.loggit.warning(
                        'Removing alias "%s" from IndexList.index_info', alias
                    )
                    del self.index_info[alias]
        self.loggit.debug('Adding "%s" to IndexList.indices', index)
        self.indices.append(index)
        self.loggit.debug(
            'Adding preliminary metadata for "%s" to IndexList.index_info', index
        )
        self.__build_index_info(index)
        self.loggit.debug('END mitigate_alias')

    def alias_index_check(self, data):
        """
        Check each index in data to see if it's an alias.
        """
        # self.loggit.debug('BEGIN alias_index_check')
        working_list = data[:]
        for entry in working_list:
            if self.client.indices.exists_alias(name=entry):
                index = list(self.client.indices.get_alias(name=entry).keys())[0]
                self.loggit.warning(
                    '"%s" is actually an alias for index "%s"', entry, index
                )
                self.mitigate_alias(index)
                # The mitigate_alias step ensures that the class ivars are handled
                # properly. The following ensure that we pass back a modified list
                data.remove(entry)
                data.append(index)
        # self.loggit.debug('END alias_index_check')
        return data

    def indices_exist(self, data, exec_func):
        """Check if indices exist. If one doesn't, remove it. Loop until all exist"""
        self.loggit.debug('BEGIN indices_exist')
        checking = True
        working_list = {}
        verified_data = self.alias_index_check(data)
        while checking:
            try:
                working_list.update(exec_func(verified_data))
            except NotFoundError as err:
                data.remove(self.__remove_missing(err))
                continue
            except TransportError as err:
                if '413' in err.errors:
                    msg = (
                        'Huge Payload 413 Err - Trying to get information via '
                        'multiple requests'
                    )
                    self.loggit.debug(msg)
                    working_list.update(self._bulk_queries(verified_data, exec_func))
            checking = False
        # self.loggit.debug('END indices_exist')
        return working_list

    def data_getter(self, data, exec_func):
        """
        Function that prevents unnecessary code repetition for different data
        getter methods
        """
        self.loggit.debug('BEGIN data_getter')
        checking = True
        while checking:
            working_list = self.indices_exist(data, exec_func)
            if working_list:
                for index in list(working_list.keys()):
                    try:
                        sii = self.index_info[index]
                    except KeyError:
                        self.loggit.warning(
                            'Index %s was not present at IndexList initialization, '
                            ' and may be behind an alias',
                            index,
                        )
                        self.mitigate_alias(index)
                        sii = self.index_info[index]
                        working_list = {}
                        try:
                            working_list.update(self._bulk_queries(data, exec_func))
                        except NotFoundError as err:
                            data.remove(self.__remove_missing(err))
                            continue
                    yield sii, working_list[index], index
            checking = False
        # self.loggit.debug('END data_getter')

    def population_check(self, index, key):
        """Verify that key is in self.index_info[index], and that it is populated"""
        retval = True
        # self.loggit.debug('BEGIN population_check')
        # self.loggit.debug('population_check: %s, %s', index, key)
        if index not in self.index_info:
            # This is just in case the index was somehow not populated
            self.__build_index_info(index)
        if key not in self.index_info[index]:
            self.index_info[index][key] = self.__zero_values()[key]
        if self.index_info[index][key] == self.__zero_values()[key]:
            retval = False
        # self.loggit.debug('END population_check')
        return retval

    def needs_data(self, indices, fields):
        """Check for data population in self.index_info"""
        self.loggit.debug('Indices: %s, Fields: %s', indices, fields)
        needful = []
        working_list = self.indices_exist(indices, self._get_indices_settings)
        for idx in working_list:
            count = 0
            for field in fields:
                # If the return value is True for this field, it means it's populated
                if self.population_check(idx, field):
                    count += 1
            if count == 0:
                # All values are the default/zero
                needful.append(idx)
        if fields == ['state']:
            self.loggit.debug('Always check open/close for all passed indices')
            needful = list(working_list.keys())
        self.loggit.debug('These indices need data in index_info: %s', needful)
        return needful

    def get_index_settings(self):
        """
        For each index in self.indices, populate ``index_info`` with:
            creation_date
            number_of_replicas
            number_of_shards
            routing information (if present)
        """
        self.loggit.debug('Getting index settings -- BEGIN')
        self.empty_list_check()
        fields = ['age', 'number_of_replicas', 'number_of_shards', 'routing']
        for lst in chunk_index_list(self.indices):
            # This portion here is to ensure that we're not polling for data
            # unless we must
            needful = self.needs_data(lst, fields)
            if not needful:
                # All indices are populated with some data, so we can skip
                # data collection
                continue
            # Now we only need to run on the 'needful'
            for sii, wli, _ in self.data_getter(needful, self._get_indices_settings):
                sii['age']['creation_date'] = fix_epoch(
                    wli['settings']['index']['creation_date']
                )
                sii['number_of_replicas'] = wli['settings']['index'][
                    'number_of_replicas'
                ]
                sii['number_of_shards'] = wli['settings']['index']['number_of_shards']
                if 'routing' in wli['settings']['index']:
                    sii['routing'] = wli['settings']['index']['routing']
        self.loggit.debug('Getting index settings -- END')

    def get_index_state(self):
        """
        For each index in self.indices, populate ``index_info`` with:

            state (open or closed)

        from the cat API
        """
        self.loggit.debug('Getting index state -- BEGIN')
        self.empty_list_check()
        fields = ['state']
        for lst in chunk_index_list(self.indices):
            # This portion here is to ensure that we're not polling for data
            # unless we must
            needful = self.needs_data(lst, fields)
            # Checking state is _always_ needful.
            resp = self.client.cat.indices(
                index=to_csv(needful), format='json', h='index,status'
            )
            for entry in resp:
                try:
                    self.index_info[entry['index']]['state'] = entry['status']
                except KeyError:
                    self.loggit.warning(
                        'Index %s was not present at IndexList initialization, '
                        'and may be behind an alias',
                        entry['index'],
                    )
                    self.mitigate_alias(entry['index'])
                    self.index_info[entry['index']]['state'] = entry['status']
        # self.loggit.debug('Getting index state -- END')

    def get_index_stats(self):
        """
        Populate ``index_info`` with index ``size_in_bytes``,
        ``primary_size_in_bytes`` and doc count information for each index.
        """
        self.loggit.debug('Getting index stats -- BEGIN')
        self.empty_list_check()
        fields = ['size_in_bytes', 'docs', 'primary_size_in_bytes']
        # This ensures that the index state is populated
        self.get_index_state()
        # Don't populate working_list until after the get_index state as it
        # can and will remove missing indices
        working_list = self.working_list()
        for index in self.working_list():
            if self.index_info[index]['state'] == 'close':
                working_list.remove(index)
        if working_list:
            index_lists = chunk_index_list(working_list)
            for lst in index_lists:
                # This portion here is to ensure that we're not polling for data
                # unless we must
                needful = self.needs_data(lst, fields)
                if not needful:
                    # All indices are populated with some data, so we can skip
                    # data collection
                    continue
                # Now we only need to run on the 'needful'
                for sii, wli, index in self.data_getter(
                    needful, self._get_indices_stats
                ):
                    try:
                        size = wli['total']['store']['size_in_bytes']
                        docs = wli['total']['docs']['count']
                        primary_size = wli['primaries']['store']['size_in_bytes']
                        msg = (
                            f'Index: {index}  Size: {byte_size(size)}  Docs: {docs} '
                            f'PrimarySize: {byte_size(primary_size)}'
                        )
                        self.loggit.debug(msg)
                        sii['size_in_bytes'] = size
                        sii['docs'] = docs
                        sii['primary_size_in_bytes'] = primary_size
                    except KeyError:
                        msg = f'Index stats missing for "{index}" -- might be closed'
                        self.loggit.warning(msg)
        # self.loggit.debug('Getting index stats -- END')

    def get_segment_counts(self):
        """
        Populate ``index_info`` with segment information for each index.
        """
        self.loggit.debug('Getting index segment counts')
        self.empty_list_check()
        for lst in chunk_index_list(self.indices):
            for sii, wli, _ in self.data_getter(lst, self._get_indices_segments):
                shards = wli['shards']
                segmentcount = 0
                for shardnum in shards:
                    for shard in range(0, len(shards[shardnum])):
                        segmentcount += shards[shardnum][shard]['num_search_segments']
                sii['segments'] = segmentcount

    def empty_list_check(self):
        """Raise :py:exc:`~.curator.exceptions.NoIndices` if ``indices`` is empty"""
        self.loggit.debug('Checking for empty list')
        if not self.indices:
            raise NoIndices('index_list object is empty.')

    def working_list(self):
        """
        Return the current value of ``indices`` as copy-by-value to prevent list
        stomping during iterations
        """
        # Copy by value, rather than reference to prevent list stomping during
        # iterations
        self.loggit.debug('Generating working list of indices')
        return self.indices[:]

    def _get_name_based_ages(self, timestring):
        """
        Add indices to ``index_info`` based on the age as indicated by the index
        name pattern, if it matches ``timestring``

        :param timestring: An :py:func:`time.strftime` pattern
        """
        # Check for empty list before proceeding here to prevent non-iterable condition
        self.loggit.debug('Getting ages of indices by "name"')
        self.empty_list_check()
        tstr = TimestringSearch(timestring)
        for index in self.working_list():
            epoch = tstr.get_epoch(index)
            if isinstance(epoch, int):
                self.index_info[index]['age']['name'] = epoch
            else:
                msg = (
                    f'Timestring {timestring} was not found in index {index}. '
                    f'Removing from actionable list'
                )
                self.loggit.debug(msg)
                self.indices.remove(index)

    def _get_field_stats_dates(self, field='@timestamp'):
        """
        Add indices to ``index_info`` based on the values the queries return, as
        determined by the min and max aggregated values of ``field``

        :param field: The field with the date value.  The field must be mapped in
            elasticsearch as a date datatype. Default: ``@timestamp``
        """
        self.loggit.debug('Cannot query closed indices. Omitting any closed indices.')
        self.filter_closed()
        self.loggit.debug(
            'Cannot use field_stats with empty indices. Omitting any empty indices.'
        )
        self.filter_empty()
        self.loggit.debug(
            'Getting index date by querying indices for min & max value of %s field',
            field,
        )
        self.empty_list_check()
        index_lists = chunk_index_list(self.indices)
        for lst in index_lists:
            for index in lst:
                aggs = {
                    'min': {'min': {'field': field}},
                    'max': {'max': {'field': field}},
                }
                response = self.client.search(index=index, size=0, aggs=aggs)
                self.loggit.debug('RESPONSE: %s', response)
                if response:
                    try:
                        res = response['aggregations']
                        self.loggit.debug('res: %s', res)
                        data = self.index_info[index]['age']
                        data['min_value'] = fix_epoch(res['min']['value'])
                        data['max_value'] = fix_epoch(res['max']['value'])
                        self.loggit.debug('data: %s', data)
                    except KeyError as exc:
                        raise ActionError(
                            f'Field "{field}" not found in index "{index}"'
                        ) from exc

    def _calculate_ages(
        self, source=None, timestring=None, field=None, stats_result=None
    ):
        """
        This method initiates index age calculation based on the given parameters.
        Exceptions are raised when they are improperly configured.

        Set instance variable ``age_keyfield`` for use later, if needed.

        :param source: Source of index age. Can be: ``name``, ``creation_date``,
            or ``field_stats``
        :param timestring: An :py:func:`time.strftime` string to match the datestamp
            in an index name. Only used for index filtering by ``name``.
        :param field: A timestamp field name.  Only used for ``field_stats`` based
            calculations.
        :param stats_result: Either ``min_value`` or ``max_value``.  Only used
            in conjunction with ``source=field_stats`` to choose whether to reference
            the min or max result value.
        """
        self.age_keyfield = source
        if source == 'name':
            if not timestring:
                raise MissingArgument(
                    'source "name" requires the "timestring" keyword argument'
                )
            self._get_name_based_ages(timestring)
        elif source == 'creation_date':
            # Nothing to do here as this comes from `get_settings` in __init__
            pass
        elif source == 'field_stats':
            if not field:
                raise MissingArgument(
                    'source "field_stats" requires the "field" keyword argument'
                )
            if stats_result not in ['min_value', 'max_value']:
                raise ValueError(f'Invalid value for "stats_result": {stats_result}')
            self.age_keyfield = stats_result
            self._get_field_stats_dates(field=field)
        else:
            raise ValueError(
                f'Invalid source: {source}. Must be one of "name", "creation_date", '
                f'"field_stats".'
            )

    def _sort_by_age(self, index_list, reverse=True):
        """
        Take a list of indices and sort them by date.

        By default, the youngest are first with ``reverse=True``, but the oldest
        can be first by setting ``reverse=False``
        """
        # Do the age-based sorting here.
        # Build an temporary dictionary with just index and age as the key and
        # value, respectively
        temp = {}
        for index in index_list:
            try:
                if self.index_info[index]['age'][self.age_keyfield]:
                    temp[index] = self.index_info[index]['age'][self.age_keyfield]
                else:
                    msg = (
                        f'No date for "{index}" in IndexList metadata. '
                        f'Possible timestring mismatch. Excluding index "{index}".'
                    )
                    self.__excludify(True, True, index, msg)
            except KeyError:
                msg = (
                    f'{index} does not have key "{self.age_keyfield}" in IndexList '
                    f'metadata'
                )
                self.__excludify(True, True, index, msg)
        # Sort alphabetically prior to age sort to keep sorting consistent
        temp_tuple = sorted(temp.items(), key=lambda k: k[0], reverse=reverse)
        # If reverse is True, this will sort so the youngest indices are first.
        # However, if you want oldest first, set reverse to False.
        # Effectively, this should set us up to act on everything older than
        # meets the other set criteria. It starts as a tuple, but then becomes a list.
        sorted_tuple = sorted(temp_tuple, key=lambda k: k[1], reverse=reverse)
        return [x[0] for x in sorted_tuple]

    def filter_by_regex(self, kind=None, value=None, exclude=False):
        """
        Match indices by regular expression (pattern).

        :param kind: Can be one of: ``suffix``, ``prefix``, ``regex``, or
            ``timestring``. This option defines what ``kind`` of filter you will
            be building.
        :param value: Depends on ``kind``. It is the :py:func:`time.strftime` string if
            ``kind`` is ``timestring``. It's used to build the regular expression
            for other kinds.
        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude=False``, then only matching
            indices will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices by regex')
        if kind not in ['regex', 'prefix', 'suffix', 'timestring']:
            raise ValueError(f'{kind}: Invalid value for kind')
        # Stop here if None or empty value, but zero is okay
        if value == 0:
            pass
        elif not value:
            raise ValueError(
                'Invalid None value for "value". Cannot be "None" type, empty, or False'
            )
        if kind == 'timestring':
            regex = settings.regex_map()[kind].format(get_date_regex(value))
        else:
            regex = settings.regex_map()[kind].format(value)
        self.empty_list_check()
        pattern = re.compile(regex)
        for index in self.working_list():
            self.loggit.debug('Filter by regex: Index: %s', index)
            match = pattern.search(index)
            if match:
                self.__excludify(True, exclude, index)
            else:
                self.__excludify(False, exclude, index)

    def filter_by_age(
        self,
        source='name',
        direction=None,
        timestring=None,
        unit=None,
        unit_count=None,
        field=None,
        stats_result='min_value',
        epoch=None,
        exclude=False,
        unit_count_pattern=False,
    ):
        """
        Match indices by relative age calculations.

        :param source: Source of index age. Can be one of ``name``, ``creation_date``,
            or ``field_stats``
        :param direction: Time to filter, either ``older`` or ``younger``
        :param timestring: An :py:func:`time.strftime` string to match the datestamp
            in an index name. Only used for index filtering by ``name``.
        :param unit: One of ``seconds``, ``minutes``, ``hours``, ``days``, ``weeks``,
            ``months``, or ``years``.
        :param unit_count: The count of ``unit``. ``unit_count`` * ``unit`` will
            be calculated out to the relative number of seconds.
        :param unit_count_pattern: A regular expression whose capture group identifies
            the value for ``unit_count``.
        :param field: A timestamp field name.  Only used for ``field_stats`` based
            calculations.
        :param stats_result: Either ``min_value`` or ``max_value``.  Only used
            in conjunction with ``source=field_stats`` to choose whether to reference
            the minimum or maximum result value.
        :param epoch: An epoch timestamp used in conjunction with ``unit`` and
            ``unit_count`` to establish a point of reference for calculations.
            If not provided, the current time will be used.
        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude`` is `False`, then only matching
            indices will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices by age')
        # Get timestamp point of reference, por
        por = get_point_of_reference(unit, unit_count, epoch)
        if not direction:
            raise MissingArgument('Must provide a value for "direction"')
        if direction not in ['older', 'younger']:
            raise ValueError(f'Invalid value for "direction": {direction}')
        # This filter requires index settings.
        self.get_index_settings()
        self._calculate_ages(
            source=source, timestring=timestring, field=field, stats_result=stats_result
        )
        if unit_count_pattern:
            try:
                unit_count_matcher = re.compile(unit_count_pattern)
            # pylint: disable=broad-except
            except Exception as exc:
                # We got an illegal regex, so won't be able to match anything
                self.loggit.error(
                    'Regular expression failure. Will not match unit count. Error: %s',
                    exc,
                )
                unit_count_matcher = None
        for index in self.working_list():
            try:
                remove_this_index = False
                age = int(self.index_info[index]['age'][self.age_keyfield])
                # if age == 0:
                #     msg = (
                #         f'Evaluating {index} resulted in an epoch timestamp of '
                #         f'0, meaning there is no associated date. Removing from '
                #         f'the actionable list.'
                #     )
                #     self.loggit.debug(msg)
                #     self.indices.remove(index)
                #     continue
                msg = (
                    f'Index "{index}" age ({age}), direction: "{direction}", point of '
                    f'reference, ({por})'
                )
                # Because time adds to epoch, smaller numbers are actually older
                # timestamps.
                if unit_count_pattern:
                    msg = (
                        f'unit_count_pattern is set, trying to match pattern to '
                        f'index "{index}"'
                    )
                    self.loggit.debug(msg)
                    unit_count_from_index = get_unit_count_from_name(
                        index, unit_count_matcher
                    )
                    if unit_count_from_index:
                        self.loggit.debug(
                            'Pattern matched, applying unit_count of  "%s"',
                            unit_count_from_index,
                        )
                        adjustedpor = get_point_of_reference(
                            unit, unit_count_from_index, epoch
                        )
                        msg = (
                            f'Adjusting point of reference from {por} to {adjustedpor} '
                            f'based on unit_count of {unit_count_from_index} from '
                            f'index name'
                        )
                        self.loggit.debug(msg)
                    elif unit_count == -1:
                        # Unable to match pattern and unit_count is -1, meaning no
                        # fallback, so this index is removed from the list
                        msg = (
                            f'Unable to match pattern and no fallback value set. '
                            f'Removing index "{index}" from actionable list'
                        )
                        self.loggit.debug(msg)
                        remove_this_index = True
                        adjustedpor = por
                        # necessary to avoid exception if the first index is excluded
                    else:
                        # Unable to match the pattern and unit_count is set, so
                        # fall back to using unit_count for determining whether
                        # to keep this index in the list
                        self.loggit.debug(
                            'Unable to match pattern using fallback value of "%s"',
                            unit_count,
                        )
                        adjustedpor = por
                else:
                    adjustedpor = por
                if direction == 'older':
                    agetest = age < adjustedpor
                else:
                    agetest = age > adjustedpor
                self.__excludify(agetest and not remove_this_index, exclude, index, msg)
            except KeyError:
                msg = (
                    f'Index "{index}" does not meet provided criteria. '
                    f'Removing from list.'
                )
                self.loggit.debug(msg)
                self.indices.remove(index)

    def filter_by_space(
        self,
        disk_space=None,
        reverse=True,
        use_age=False,
        source='creation_date',
        timestring=None,
        field=None,
        stats_result='min_value',
        exclude=False,
        threshold_behavior='greater_than',
    ):
        """
        Remove indices from the actionable list based on space consumed, sorted
        reverse-alphabetically by default.  If you set ``reverse`` to ``False``,
        it will be sorted alphabetically.

        The default is usually what you will want. If only one kind of index is
        provided--for example, indices matching ``logstash-%Y.%m.%d`` --then
        reverse alphabetical sorting will mean the oldest will remain in the list,
        because lower numbers in the dates mean older indices.

        By setting ``reverse`` to ``False``, then ``index3`` will be deleted before
        ``index2``, which will be deleted before ``index1``

        ``use_age`` allows ordering indices by age. Age is determined by the
        index creation date by default, but you can specify an ``source`` of
        ``name``, ``max_value``, or ``min_value``. The ``name`` ``source`` requires
        the timestring argument.

        ``threshold_behavior``, when set to ``greater_than`` (default), includes
        if it the index tests to be larger than ``disk_space``. When set to
        ``less_than``, it includes if the index is smaller than ``disk_space``

        :param disk_space: Filter indices over *n* gigabytes
        :param threshold_behavior: Size to filter, either ``greater_than`` or
            ``less_than``. Defaults to ``greater_than`` to preserve backwards
            compatability.
        :param reverse: The filtering direction. (default: ``True``).  Ignored if
            ``use_age`` is ``True``
        :param use_age: Sort indices by age.  ``source`` is required in this case.
        :param source: Source of index age. Can be one of ``name``, ``creation_date``,
            or ``field_stats``. Default: ``creation_date``
        :param timestring: An :py:func:`time.strftime` string to match the datestamp
            in an index name. Only used if ``source=name`` is selected.
        :param field: A timestamp field name.  Only used if ``source=field_stats``
            is selected.
        :param stats_result: Either ``min_value`` or ``max_value``.  Only used if
            ``source=field_stats`` is selected. It determines whether to reference
            the minimum or maximum value of `field` in each index.
        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude=False``, then only matching
            indices will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices by disk space')
        # Ensure that disk_space is a float
        if not disk_space:
            raise MissingArgument('No value for "disk_space" provided')
        if threshold_behavior not in ['greater_than', 'less_than']:
            raise ValueError(
                f'Invalid value for "threshold_behavior": {threshold_behavior}'
            )
        # This filter requires both index stats and index settings
        self.get_index_stats()
        self.get_index_settings()
        disk_space = float(disk_space)
        disk_usage = 0.0
        disk_limit = disk_space * 2**30
        msg = (
            'Cannot get disk usage info from closed indices. Omitting any '
            'closed indices.'
        )
        self.loggit.debug(msg)
        self.filter_closed()
        if use_age:
            self._calculate_ages(
                source=source,
                timestring=timestring,
                field=field,
                stats_result=stats_result,
            )
            # Using default value of reverse=True in self._sort_by_age()
            self.loggit.debug('SORTING BY AGE')
            sorted_indices = self._sort_by_age(self.working_list())
        else:
            # Default to sorting by index name
            sorted_indices = sorted(self.working_list(), reverse=reverse)
        for index in sorted_indices:
            disk_usage += self.index_info[index]['size_in_bytes']
            msg = (
                f'{index}, summed disk usage is {byte_size(disk_usage)} and disk limit '
                f'is {byte_size(disk_limit)}.'
            )
            if threshold_behavior == 'greater_than':
                self.__excludify((disk_usage > disk_limit), exclude, index, msg)
            elif threshold_behavior == 'less_than':
                self.__excludify((disk_usage < disk_limit), exclude, index, msg)

    def filter_kibana(self, exclude=True):
        """
        Match any index named ``.kibana*`` in ``indices``. Older releases addressed
        index names that no longer exist.

        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude=False``, then only matching
            indices will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering kibana indices')
        self.empty_list_check()
        for index in self.working_list():
            pattern = re.compile(r'^\.kibana.*$')
            if pattern.match(index):
                self.__excludify(True, exclude, index)
            else:
                self.__excludify(False, exclude, index)

    def filter_forceMerged(self, max_num_segments=None, exclude=True):
        """
        Match any index which has ``max_num_segments`` per shard or fewer in the
        actionable list.

        :param max_num_segments: Cutoff number of segments per shard.
        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude=False``, then only matching
            indices will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering forceMerged indices')
        if not max_num_segments:
            raise MissingArgument('Missing value for "max_num_segments"')
        self.loggit.debug(
            'Cannot get segment count of closed indices. Omitting any closed indices.'
        )
        # This filter requires the index state (open/close), and index settings.
        self.get_index_state()
        self.get_index_settings()
        self.filter_closed()
        self.get_segment_counts()
        for index in self.working_list():
            # Do this to reduce long lines and make it more readable...
            shards = int(self.index_info[index]['number_of_shards'])
            replicas = int(self.index_info[index]['number_of_replicas'])
            segments = int(self.index_info[index]['segments'])
            msg = (
                f'{index} has {shards} shard(s) + {replicas} replica(s) '
                f'with a sum total of {segments} segments.'
            )
            expected_count = (shards + (shards * replicas)) * max_num_segments
            self.__excludify((segments <= expected_count), exclude, index, msg)

    def filter_closed(self, exclude=True):
        """
        Filter out closed indices from ``indices``

        :param exclude: If ``exclude=True``, this filter will remove matching
            indices from ``indices``. If ``exclude=False``, then only matching
            indices will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering closed indices')
        # This filter requires index state (open/close)
        self.get_index_state()
        self.empty_list_check()
        for index in self.working_list():
            condition = self.index_info[index]['state'] == 'close'
            self.loggit.debug(
                'Index %s state: %s', index, self.index_info[index]['state']
            )
            self.__excludify(condition, exclude, index)

    def filter_empty(self, exclude=True):
        """
        Filter indices with a document count of zero. Indices that are closed
        are automatically excluded from consideration due to closed indices reporting
        a document count of zero.

        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering empty indices')
        # This index requires index state (open/close) and index stats
        self.get_index_state()
        self.get_index_stats()
        self.filter_closed()
        self.empty_list_check()
        for index in self.working_list():
            condition = self.index_info[index]['docs'] == 0
            self.loggit.debug(
                'Index %s doc count: %s', index, self.index_info[index]['docs']
            )
            self.__excludify(condition, exclude, index)

    def filter_opened(self, exclude=True):
        """
        Filter out opened indices from ``indices``

        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering open indices')
        # This filter requires index state (open/close)
        self.get_index_state()
        self.empty_list_check()
        for index in self.working_list():
            condition = self.index_info[index]['state'] == 'open'
            self.loggit.debug(
                'Index %s state: %s', index, self.index_info[index]['state']
            )
            self.__excludify(condition, exclude, index)

    def filter_allocated(
        self, key=None, value=None, allocation_type='require', exclude=True
    ):
        """
        Match indices that have the routing allocation rule of ``key=value`` from
        ``indices``

        :param key: The allocation attribute to check for
        :param value: The value to check for
        :param allocation_type: Type of allocation to apply
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is `T`rue`
        """
        self.loggit.debug('Filtering indices with shard routing allocation rules')
        if not key:
            raise MissingArgument('No value for "key" provided')
        if not value:
            raise MissingArgument('No value for "value" provided')
        if allocation_type not in ['include', 'exclude', 'require']:
            raise ValueError(f'Invalid "allocation_type": {allocation_type}')
        # This filter requires index settings
        self.get_index_settings()
        self.get_index_state()
        self.empty_list_check()
        for lst in chunk_index_list(self.indices):
            working_list = self._get_indices_settings(lst)
            if working_list:
                for index in list(working_list.keys()):
                    try:
                        has_routing = (
                            working_list[index]['settings']['index']['routing'][
                                'allocation'
                            ][allocation_type][key]
                            == value
                        )
                    except KeyError:
                        has_routing = False
                    # if has_routing:
                    msg = (
                        f'{index}: Routing (mis)match: '
                        f'index.routing.allocation.{allocation_type}.{key}={value}.'
                    )
                    self.__excludify(has_routing, exclude, index, msg)

    def filter_none(self):
        """The legendary NULL filter"""
        self.loggit.debug('"None" filter selected.  No filtering will be done.')

    def filter_by_alias(self, aliases=None, exclude=False):
        """
        Match indices which are associated with the alias or list of aliases
        identified by ``aliases``. Indices must appear in all aliases in list
        ``aliases`` or a 404 error will result, leading to no indices being matched.

        :param aliases: A list of alias names.
        :type aliases: list
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices matching aliases: "%s"', aliases)
        if not aliases:
            raise MissingArgument('No value for "aliases" provided')
        aliases = ensure_list(aliases)
        self.empty_list_check()
        for lst in chunk_index_list(self.indices):
            try:
                # get_alias will either return {} or a NotFoundError.
                has_alias = list(
                    self.client.indices.get_alias(
                        index=to_csv(lst), name=to_csv(aliases)
                    ).keys()
                )
                self.loggit.debug('has_alias: %s', has_alias)
            except NotFoundError:
                # if we see the NotFoundError, we need to set working_list to {}
                has_alias = []
            for index in lst:
                if index in has_alias:
                    isness = 'is'
                    condition = True
                else:
                    isness = 'is not'
                    condition = False
                msg = f'{index} {isness} associated with aliases: {aliases}'
                self.__excludify(condition, exclude, index, msg)

    def filter_by_count(
        self,
        count=None,
        reverse=True,
        use_age=False,
        pattern=None,
        source='creation_date',
        timestring=None,
        field=None,
        stats_result='min_value',
        exclude=True,
    ):
        """
        Remove indices from the actionable list beyond the number ``count``, sorted
        reverse-alphabetically by default.  If you set ``reverse=False``, it will
        be sorted alphabetically.

        The default is usually what you will want. If only one kind of index is
        provided--for example, indices matching ``logstash-%Y.%m.%d`` -- then
        reverse alphabetical sorting will mean the oldest will remain in the list,
        because lower numbers in the dates mean older indices.

        By setting ``reverse=False``, then ``index3`` will be deleted before
        ``index2``, which will be deleted before ``index1``

        ``use_age`` allows ordering indices by age. Age is determined by the index
        creation date by default, but you can specify an ``source`` of ``name``,
        ``max_value``, or ``min_value``. The ``name`` `source` requires the
        timestring argument.

        :param count: Filter indices beyond ``count``.
        :param reverse: The filtering direction. (default: ``True``).
        :param use_age: Sort indices by age.  ``source`` is required in this case.
        :param pattern: Select indices to count from a regular expression pattern.
            This pattern must have one and only one capture group. This can allow
            a single ``count`` filter instance to operate against any number of
            matching patterns, and keep ``count`` of each index in that group.
            For example, given a ``pattern`` of ``'^(.*)-\\d{6}$'``, it will match
            both ``rollover-000001`` and ``index-999990``, but not
            ``logstash-2017.10.12``. Following the same example, if my cluster
            also had ``rollover-000002`` through ``rollover-000010`` and
            ``index-888888`` through ``index-999999``, it will process both groups
            of indices, and include or exclude the ``count`` of each.
        :param source: Source of index age. Can be one of ``name``,
            ``creation_date``, or ``field_stats``. Default: ``creation_date``
        :param timestring: An :py:func:`time.strftime` string to match the datestamp
            in an index name. Only used if ``source=name``.
        :param field: A timestamp field name.  Only used if ``source=field_stats``.
        :param stats_result: Either ``min_value`` or ``max_value``.  Only used if
            ``source=field_stats``. It determines whether to reference the minimum
            or maximum value of ``field`` in each index.
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering indices by count')
        if not count:
            raise MissingArgument('No value for "count" provided')
        # This filter requires index state (open/close) and index settings
        self.get_index_state()
        self.get_index_settings()
        # Create a copy-by-value working list
        working_list = self.working_list()
        if pattern:
            try:
                regex = re.compile(pattern)
                if regex.groups < 1:
                    raise ConfigurationError(
                        f'No regular expression group found in {pattern}'
                    )
                if regex.groups > 1:
                    raise ConfigurationError(
                        f'More than 1 regular expression group found in {pattern}'
                    )
                # Prune indices not matching the regular expression the object
                # (And filtered_indices) We do not want to act on them by accident.
                prune_these = list(
                    filter(lambda x: regex.match(x) is None, working_list)
                )
                filtered_indices = working_list
                for index in prune_these:
                    msg = '{index} does not match regular expression {pattern}.'
                    condition = True
                    exclude = True
                    self.__excludify(condition, exclude, index, msg)
                    # also remove it from filtered_indices
                    filtered_indices.remove(index)
                # Presort these filtered_indices using the lambda
                presorted = sorted(
                    filtered_indices, key=lambda x: regex.match(x).group(1)
                )
            except Exception as exc:
                raise ActionError(
                    f'Unable to process pattern: "{pattern}". Error: {exc}'
                ) from exc
            # Initialize groups here
            groups = []
            # We have to pull keys k this way, but we don't need to keep them
            # We only need g for groups
            for _, g in itertools.groupby(
                presorted, key=lambda x: regex.match(x).group(1)
            ):
                groups.append(list(g))
        else:
            # Since pattern will create a list of lists, and we iterate over that,
            # we need to put our single list inside a list
            groups = [working_list]
        for group in groups:
            if use_age:
                if source != 'name':
                    self.loggit.warning(
                        'Cannot get age information from closed indices unless '
                        'source="name".  Omitting any closed indices.'
                    )
                    self.filter_closed()
                self._calculate_ages(
                    source=source,
                    timestring=timestring,
                    field=field,
                    stats_result=stats_result,
                )
                # Using default value of reverse=True in self._sort_by_age()
                sorted_indices = self._sort_by_age(group, reverse=reverse)
            else:
                # Default to sorting by index name
                sorted_indices = sorted(group, reverse=reverse)
            idx = 1
            for index in sorted_indices:
                msg = f'{index} is {idx} of specified count of {count}.'
                condition = True if idx <= count else False
                self.__excludify(condition, exclude, index, msg)
                idx += 1

    def filter_by_shards(
        self, number_of_shards=None, shard_filter_behavior='greater_than', exclude=False
    ):
        """
        Match ``indices`` with a given shard count.

        Selects all indices with a shard count ``greater_than`` ``number_of_shards``
        by default. Use ``shard_filter_behavior`` to select indices with shard
        count ``greater_than``, ``greater_than_or_equal``, ``less_than``,
        ``less_than_or_equal``, or ``equal`` to ``number_of_shards``.

        :param number_of_shards: shard threshold
        :param shard_filter_behavior: Do you want to filter on ``greater_than``,
            ``greater_than_or_equal``, ``less_than``, ``less_than_or_equal``,
            or ``equal``?
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug("Filtering indices by number of shards")
        if not number_of_shards:
            raise MissingArgument('No value for "number_of_shards" provided')
        if shard_filter_behavior not in [
            'greater_than',
            'less_than',
            'greater_than_or_equal',
            'less_than_or_equal',
            'equal',
        ]:
            raise ValueError(
                f'Invalid value for "shard_filter_behavior": {shard_filter_behavior}'
            )
        if number_of_shards < 1 or (
            shard_filter_behavior == 'less_than' and number_of_shards == 1
        ):
            raise ValueError(
                f'Unacceptable value: {number_of_shards} -- "number_of_shards" cannot '
                f'be less than 1. A valid index will have at least one shard.'
            )
        # This filter requires index_settings to count shards
        self.get_index_settings()
        self.empty_list_check()
        for index in self.working_list():
            self.loggit.debug('Filter by number of shards: Index: %s', index)
            if shard_filter_behavior == 'greater_than':
                condition = (
                    int(self.index_info[index]['number_of_shards']) > number_of_shards
                )
            elif shard_filter_behavior == 'less_than':
                condition = (
                    int(self.index_info[index]['number_of_shards']) < number_of_shards
                )
            elif shard_filter_behavior == 'greater_than_or_equal':
                condition = (
                    int(self.index_info[index]['number_of_shards']) >= number_of_shards
                )
            elif shard_filter_behavior == 'less_than_or_equal':
                condition = (
                    int(self.index_info[index]['number_of_shards']) <= number_of_shards
                )
            else:
                condition = (
                    int(self.index_info[index]['number_of_shards']) == number_of_shards
                )
            self.__excludify(condition, exclude, index)

    def filter_period(
        self,
        period_type='relative',
        source='name',
        range_from=None,
        range_to=None,
        date_from=None,
        date_to=None,
        date_from_format=None,
        date_to_format=None,
        timestring=None,
        unit=None,
        field=None,
        stats_result='min_value',
        intersect=False,
        week_starts_on='sunday',
        epoch=None,
        exclude=False,
    ):
        """
        Match ``indices`` with ages within a given period.

        :param period_type: Can be either ``absolute`` or ``relative``.  Default
            is ``relative``. ``date_from`` and ``date_to`` are required when using
            ``period_type='absolute'``. ``range_from`` and ``range_to`` are required
            with ``period_type='relative'``.
        :param source: Source of index age. Can be ``name``, ``creation_date``,
            or ``field_stats``
        :param range_from: How many ``unit`` (s) in the past/future is the origin?
        :param range_to: How many ``unit`` (s) in the past/future is the end point?
        :param date_from: The simplified date for the start of the range
        :param date_to: The simplified date for the end of the range.  If this
            value is the same as ``date_from``, the full value of ``unit`` will be
            extrapolated for the range.  For example, if ``unit=months``, and
            ``date_from`` and ``date_to`` are both ``2017.01``, then the entire
            month of January 2017 will be the absolute date range.
        :param date_from_format: The :py:func:`time.strftime` string used to
            parse ``date_from``
        :param date_to_format: The :py:func:`time.strftime` string used to
            parse ``date_to``
        :param timestring: An :py:func:`time.strftime` string to match the datestamp
            in an index name. Only used for index filtering by ``name``.
        :param unit: One of ``hours``, ``days``, ``weeks``, ``months``, or ``years``.
        :param field: A timestamp field name.  Only used for ``field_stats`` based
            calculations.
        :param stats_result: Either ``min_value`` or ``max_value``.  Only used in
            conjunction with ``source='field_stats'`` to choose whether to reference
            the min or max result value.
        :param intersect: Only used when ``source='field_stats'``. If ``True``,
            only indices where both ``min_value`` and ``max_value`` are within the
            period will be selected. If ``False``, it will use whichever you specified.
            Default is ``False`` to preserve expected behavior.
        :param week_starts_on: Either ``sunday`` or ``monday``. Default is ``sunday``
        :param epoch: An epoch timestamp used to establish a point of reference for
            calculations. If not provided, the current time will be used.
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices by period')
        if period_type not in ['absolute', 'relative']:
            raise ValueError(
                f'Unacceptable value: {period_type} -- "period_type" must be either '
                f'"absolute" or "relative".'
            )
        if period_type == 'relative':
            func = date_range
            args = [unit, range_from, range_to, epoch]
            kwgs = {'week_starts_on': week_starts_on}
            if (not isinstance(range_from, int)) or (not isinstance(range_to, int)):
                raise ConfigurationError(
                    '"range_from" and "range_to" must be integer values'
                )
        else:
            func = absolute_date_range
            args = [unit, date_from, date_to]
            kwgs = {
                'date_from_format': date_from_format,
                'date_to_format': date_to_format,
            }
            for reqd in [date_from, date_to, date_from_format, date_to_format]:
                if not reqd:
                    raise ConfigurationError(
                        'Must provide "date_from", "date_to", "date_from_format", and '
                        '"date_to_format" with absolute period_type'
                    )
        # This filter requires index settings
        self.get_index_settings()
        try:
            start, end = func(*args, **kwgs)
        # pylint: disable=broad-except
        except Exception as exc:
            report_failure(exc)
        self._calculate_ages(
            source=source, timestring=timestring, field=field, stats_result=stats_result
        )
        for index in self.working_list():
            try:
                if source == 'field_stats' and intersect:
                    min_age = int(self.index_info[index]['age']['min_value'])
                    max_age = int(self.index_info[index]['age']['max_value'])
                    msg = (
                        f'Index "{index}", timestamp field "{field}", min_value '
                        f'({min_age}), max_value ({max_age}), period start: '
                        f'"{start}", period end, "{end}"'
                    )
                    # Because time adds to epoch, smaller numbers are actually older
                    # timestamps.
                    inrange = (min_age >= start) and (max_age <= end)
                else:
                    age = int(self.index_info[index]['age'][self.age_keyfield])
                    msg = (
                        f'Index "{index}" age ({age}), period start: "{start}", period '
                        f'end, "{end}"'
                    )
                    # Because time adds to epoch, smaller numbers are actually older
                    # timestamps.
                    inrange = (age >= start) and (age <= end)
                self.__excludify(inrange, exclude, index, msg)
            except KeyError:
                self.loggit.debug(
                    'Index "%s" does not meet provided criteria. Removing from list.',
                    index,
                )
                self.indices.remove(index)

    def filter_ilm(self, exclude=True):
        """
        Match indices that have the setting ``index.lifecycle.name``

        :param exclude: If ``exclude=True``, this filter will remove matching
            ``indices``. If ``exclude=False``, then only matching indices will be
            kept in ``indices``. Default is ``True``
        """
        self.loggit.debug('Filtering indices with index.lifecycle.name')
        index_lists = chunk_index_list(self.indices)
        if index_lists == [['']]:
            self.loggit.debug('Empty working list. No ILM indices to filter.')
            return
        for lst in index_lists:
            working_list = self._get_indices_settings(lst)
            if working_list:
                for index in list(working_list.keys()):
                    try:
                        subvalue = working_list[index]['settings']['index']['lifecycle']
                        has_ilm = 'name' in subvalue
                        msg = f"{index} has index.lifecycle.name {subvalue['name']}"
                    except KeyError:
                        has_ilm = False
                        msg = f'index.lifecycle.name is not set for index {index}'
                    self.__excludify(has_ilm, exclude, index, msg)

    def iterate_filters(self, filter_dict):
        """
        Iterate over the filters defined in ``config`` and execute them.

        :param filter_dict: The configuration dictionary

        .. note:: ``filter_dict`` should be a dictionary with the following form:
        .. code-block:: python

                { 'filters' : [
                        {
                            'filtertype': 'the_filter_type',
                            'key1' : 'value1',
                            ...
                            'keyN' : 'valueN'
                        }
                    ]
                }

        """
        self.loggit.debug('Iterating over a list of filters')
        # Make sure we actually _have_ filters to act on
        if 'filters' not in filter_dict or len(filter_dict['filters']) < 1:
            self.loggit.info('No filters in config.  Returning unaltered object.')
            return
        self.loggit.debug('All filters: %s', filter_dict['filters'])
        for fil in filter_dict['filters']:
            self.loggit.debug('Top of the loop: %s', self.indices)
            self.loggit.debug('Un-parsed filter args: %s', fil)
            # Make sure we got at least this much in the configuration
            chk = SchemaCheck(
                fil, filterstructure(), 'filter', 'IndexList.iterate_filters'
            ).result()
            msg = f'Parsed filter args: {chk}'
            self.loggit.debug(msg)
            method = self.__map_method(fil['filtertype'])
            del fil['filtertype']
            # If it's a filtertype with arguments, update the defaults with the
            # provided settings.
            if fil:
                self.loggit.debug('Filter args: %s', fil)
                self.loggit.debug('Pre-instance: %s', self.indices)
                method(**fil)
                self.loggit.debug('Post-instance: %s', self.indices)
            else:
                # Otherwise, it's a settingless filter.
                method()

    def filter_by_size(
        self,
        size_threshold=None,
        threshold_behavior='greater_than',
        exclude=False,
        size_behavior='primary',
    ):
        """
        Remove indices from the actionable list based on index size.

        ``threshold_behavior``, when set to ``greater_than`` (default), includes
        if it the index tests to be larger than ``size_threshold``. When set to
        ``less_than``, it includes if the index is smaller than ``size_threshold``

        :param size_threshold: Filter indices over *n* gigabytes
        :param threshold_behavior: Size to filter, either ``greater_than`` or
            ``less_than``. Defaults to ``greater_than`` to preserve backwards
            compatability.
        :param size_behavior: Size that used to filter, either ``primary`` or
            ``total``. Defaults to ``primary``
        :param exclude: If ``exclude=True``, this filter will remove matching indices
            from ``indices``. If ``exclude=False``, then only matching indices
            will be kept in ``indices``. Default is ``False``
        """
        self.loggit.debug('Filtering indices by index size')
        # Ensure that disk_space is a float
        if not size_threshold:
            raise MissingArgument('No value for "size_threshold" provided')
        if size_behavior not in ['primary', 'total']:
            raise ValueError(f'Invalid value for "size_behavior": {size_behavior}')
        if threshold_behavior not in ['greater_than', 'less_than']:
            raise ValueError(
                f'Invalid value for "threshold_behavior": {threshold_behavior}'
            )
        index_size_limit = float(size_threshold) * 2**30
        msg = (
            'Cannot get disk usage info from closed indices. '
            'Omitting any closed indices.'
        )
        self.loggit.debug(msg)
        # This filter requires index state (open/close) and index stats
        self.get_index_state()
        self.get_index_stats()
        self.filter_closed()
        # Create a copy-by-value working list
        working_list = self.working_list()
        for index in working_list:
            if size_behavior == 'primary':
                index_size = self.index_info[index]['primary_size_in_bytes']
            else:
                index_size = self.index_info[index]['size_in_bytes']
            msg = (
                f'{index}, index size is {byte_size(index_size)} and '
                f'size limit is {byte_size(index_size_limit)}.'
            )
            if threshold_behavior == 'greater_than':
                self.__excludify((index_size > index_size_limit), exclude, index, msg)
            elif threshold_behavior == 'less_than':
                self.__excludify((index_size < index_size_limit), exclude, index, msg)



"""Validate root ``actions`` and individual ``action`` Schemas"""

from voluptuous import Any, In, Schema, Optional, Required
from es_client.helpers.schemacheck import SchemaCheck
from curator.defaults import settings


def root():
    """
    Return a valid :py:class:`~.voluptuous.schema_builder.Schema` definition which
    is a dictionary with ``actions`` :py:class:`~.voluptuous.schema_builder.Required`
    to be the root key with another dictionary as the value.
    """
    return Schema({Required('actions'): dict})


def valid_action():
    """
    Return a valid :py:class:`~.voluptuous.schema_builder.Schema` definition which is
    that the value of key ``action`` is
    :py:class:`~.voluptuous.schema_builder.Required` to be
    :py:class:`~.voluptuous.schema_builder.In` the value returned by
    :py:func:`~.curator.defaults.settings.all_actions`.
    """
    return {
        Required('action'): Any(
            In(settings.all_actions()),
            msg=f'action must be one of {settings.all_actions()}',
        )
    }


def structure(data, location):
    """
    Return a valid :py:class:`~.voluptuous.schema_builder.Schema` definition which
    tests ``data``, which is ostensibly an individual action dictionary. If it is a
    :py:func:`~.curator.validators.actions.valid_action`, then it will
    :py:meth:`~.voluptuous.schema_builder.Schema.update` the base
    :py:class:`~.voluptuous.schema_builder.Schema` with other options, based on the
    what the value of ``data['action']`` is.

    :param data: The configuration dictionary, or sub-dictionary, being validated
    :type data: dict
    :param location: A string to report which configuration sub-block is being tested.
    :type location: str
    :returns: A :py:class:`~.voluptuous.schema_builder.Schema` object
    """
    _ = SchemaCheck(
        data,
        Schema(valid_action(), extra=True),
        'action type',
        location,
    ).result()
    # Build a valid schema knowing that the action has already been validated
    retval = valid_action()
    retval.update({Optional('description', default='No description given'): Any(str)})
    retval.update({Optional('options', default=settings.default_options()): dict})
    action = data['action']
    if action in ['cluster_routing', 'create_index', 'rollover']:
        # The cluster_routing, create_index, and rollover actions should not
        # have a 'filters' block
        pass
    elif action == 'alias':
        # The alias action should not have a filters block, but should have
        # an add and/or remove block.
        retval.update(
            {
                Optional('add'): dict,
                Optional('remove'): dict,
            }
        )
    else:
        retval.update({Optional('filters', default=settings.default_filters()): list})
    return Schema(retval)



"""Index settings action class"""

import logging

# pylint: disable=import-error
from curator.exceptions import ActionError, ConfigurationError, MissingArgument
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure, show_dry_run, to_csv


class IndexSettings:
    """Index Settings Action Class"""

    def __init__(
        self,
        ilo,
        index_settings=None,
        ignore_unavailable=False,
        preserve_existing=False,
    ):
        """
        :param ilo: An IndexList Object
        :param index_settings: A settings structure with one or more index
            settings to change.
        :param ignore_unavailable: Whether specified concrete indices should be
            ignored when unavailable (missing or closed)
        :param preserve_existing: Whether to update existing settings. If set to
            ``True``, existing settings on an index remain unchanged. The default
            is ``False``

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type index_settings: dict
        :type ignore_unavailable: bool
        :type preserve_existing: bool
        """
        if index_settings is None:
            index_settings = {}
        verify_index_list(ilo)
        if not index_settings:
            raise MissingArgument('Missing value for "index_settings"')
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from
        #: param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: Object attribute that gets the value of param ``index_settings``.
        self.body = index_settings
        #: Object attribute that gets the value of param ``ignore_unavailable``.
        self.ignore_unavailable = ignore_unavailable
        #: Object attribute that gets the value of param ``preserve_existing``.
        self.preserve_existing = preserve_existing

        self.loggit = logging.getLogger('curator.actions.index_settings')
        self._body_check()

    def _body_check(self):
        # The body only passes the skimpiest of requirements by having 'index'
        # as the only root-level key, and having a 'dict' as its value
        if len(self.body) == 1:
            if 'index' in self.body:
                if isinstance(self.body['index'], dict):
                    return True
        raise ConfigurationError(f'Bad value for "index_settings": {self.body}')

    def _static_settings(self):
        return [
            'number_of_shards',
            'shard',
            'codec',
            'routing_partition_size',
        ]

    def _dynamic_settings(self):
        return [
            'number_of_replicas',
            'auto_expand_replicas',
            'refresh_interval',
            'max_result_window',
            'max_rescore_window',
            'blocks',
            'max_refresh_listeners',
            'mapping',
            'merge',
            'translog',
        ]

    def _settings_check(self):
        # Detect if even one index is open.  Save all found to open_index_list.
        open_index_list = []
        open_indices = False
        # This action requires index settings and state to be present
        # Calling these here should not cause undue problems, even if it's a repeat call
        self.index_list.get_index_state()
        self.index_list.get_index_settings()
        for idx in self.index_list.indices:
            if self.index_list.index_info[idx]['state'] == 'open':
                open_index_list.append(idx)
                open_indices = True
        for k in self.body['index']:
            if k in self._static_settings():
                if not self.ignore_unavailable:
                    if open_indices:
                        msg = (
                            f'Static Setting "{k}" detected with open indices: '
                            f'{open_index_list}. Static settings can only be used '
                            f'with closed indices.  Recommend filtering out open '
                            f'indices, or setting ignore_unavailable to True'
                        )
                        raise ActionError(msg)
            elif k in self._dynamic_settings():
                # Dynamic settings should be appliable to open or closed indices
                # Act here if the case is different for some settings.
                pass
            else:
                msg = (
                    f'"{k}" is not a setting Curator recognizes and may or may '
                    f' not work.'
                )
                self.loggit.warning(msg)

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        show_dry_run(self.index_list, 'indexsettings', **self.body)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.put_settings` in :py:attr:`body`
        to indices in :py:attr:`index_list`
        """
        self._settings_check()
        # Ensure that the open indices filter applied in _settings_check()
        # didn't result in an empty list (or otherwise empty)
        self.index_list.empty_list_check()
        msg = (
            f'Applying index settings to {len(self.index_list.indices)} indices: '
            f'{self.index_list.indices}'
        )
        self.loggit.info(msg)
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                response = self.client.indices.put_settings(
                    index=to_csv(lst),
                    body=self.body,
                    ignore_unavailable=self.ignore_unavailable,
                    preserve_existing=self.preserve_existing,
                )
                self.loggit.debug('PUT SETTINGS RESPONSE: %s', response)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Object builder"""

import logging
import sys
from voluptuous import Schema
from es_client.builder import Builder
from es_client.exceptions import FailedValidation
from es_client.helpers.schemacheck import SchemaCheck
from es_client.helpers.utils import prune_nones
from curator import IndexList, SnapshotList
from curator.actions import (
    Alias,
    Allocation,
    Close,
    ClusterRouting,
    CreateIndex,
    DeleteIndices,
    ForceMerge,
    IndexSettings,
    Open,
    Reindex,
    Replicas,
    Rollover,
    Shrink,
    Snapshot,
    DeleteSnapshots,
    Restore,
)
from curator.defaults.settings import VERSION_MAX, VERSION_MIN, snapshot_actions
from curator.exceptions import ConfigurationError, NoIndices, NoSnapshots
from curator.helpers.testers import validate_filters
from curator.validators import options
from curator.validators.filter_functions import validfilters

CLASS_MAP = {
    'alias': Alias,
    'allocation': Allocation,
    'close': Close,
    'cluster_routing': ClusterRouting,
    'create_index': CreateIndex,
    'delete_indices': DeleteIndices,
    'delete_snapshots': DeleteSnapshots,
    'forcemerge': ForceMerge,
    'index_settings': IndexSettings,
    'open': Open,
    'reindex': Reindex,
    'replicas': Replicas,
    'restore': Restore,
    'rollover': Rollover,
    'shrink': Shrink,
    'snapshot': Snapshot,
}

EXCLUDED_OPTIONS = [
    'ignore_empty_list',
    'timeout_override',
    'continue_if_exception',
    'disable_action',
]


class CLIAction:
    """
    Unified class for all CLI singleton actions
    """

    def __init__(
        self, action, configdict, option_dict, filter_list, ignore_empty_list, **kwargs
    ):
        """Class setup
        :param action: The action name.
        :param configdict: ``dict`` containing everything needed for
            :py:class:`~.es_client.builder.Builder` to build an
            :py:class:`~.elasticsearch.Elasticsearch` client object.
        :param option_dict: Options for ``action``.
        :param filter_list: Filters to select indices for ``action``.
        :param ignore_empty_list: Exit ``0`` even if filters result in no indices
            for ``action``.
        :param kwargs: Other keyword args to pass to ``action``.

        :type action: str
        :type configdict: dict
        :type option_dict: dict
        :type filter_list: list
        :type ignore_empty_list: bool
        :type kwargs: dict
        """
        self.logger = logging.getLogger('curator.cli_singletons.cli_action.' + action)
        self.filters = []
        self.action = action
        self.list_object = None
        self.repository = kwargs['repository'] if 'repository' in kwargs else None
        if action[:5] != 'show_':  # Ignore CLASS_MAP for show_indices/show_snapshots
            try:
                self.action_class = CLASS_MAP[action]
            except KeyError:
                self.logger.critical('Action must be one of %s', list(CLASS_MAP.keys()))
            self.check_options(option_dict)
        else:
            self.options = option_dict

        self.search_pattern = self.options.pop('search_pattern', '*')

        # Extract allow_ilm_indices so it can be handled separately.
        if 'allow_ilm_indices' in self.options:
            self.allow_ilm = self.options.pop('allow_ilm_indices')
        else:
            self.allow_ilm = False
        if action == 'alias':
            self.logger.debug('ACTION = ALIAS')
            self.alias = {
                'name': option_dict['name'],
                'extra_settings': option_dict['extra_settings'],
                'wini': (
                    kwargs['warn_if_no_indices']
                    if 'warn_if_no_indices' in kwargs
                    else False
                ),
            }
            for k in ['add', 'remove']:
                if k in kwargs:
                    self.alias[k] = {}
                    self.check_filters(kwargs[k], loc='alias singleton', key=k)
                    self.alias[k]['filters'] = self.filters
                    if self.allow_ilm:
                        self.alias[k]['filters'].append({'filtertype': 'ilm'})
        # No filters for these actions
        elif action in ['cluster_routing', 'create_index', 'rollover']:
            self.action_kwargs = {}
            if action == 'rollover':
                self.logger.debug('rollover option_dict = %s', option_dict)
        else:
            self.check_filters(filter_list)
        try:
            builder = Builder(
                configdict=configdict, version_max=VERSION_MAX, version_min=VERSION_MIN
            )
            builder.connect()
        # pylint: disable=broad-except
        except Exception as exc:
            raise ConfigurationError(
                f'Unable to connect to Elasticsearch as configured: {exc}'
            ) from exc
        # If we're here, we'll see the output from GET http(s)://hostname.tld:PORT
        self.logger.debug('Connection result: %s', builder.client.info())
        self.client = builder.client
        self.ignore = ignore_empty_list

    def prune_excluded(self, option_dict):
        """Prune excluded options"""
        for k in list(option_dict.keys()):
            if k in EXCLUDED_OPTIONS:
                del option_dict[k]
        return option_dict

    def check_options(self, option_dict):
        """Validate provided options"""
        try:
            self.logger.debug('Validating provided options: %s', option_dict)
            # Kludgy work-around to needing 'repository' in options for these actions
            # but only to pass the schema check.  It's removed again below.
            if self.action in ['delete_snapshots', 'restore']:
                option_dict['repository'] = self.repository
            _ = SchemaCheck(
                prune_nones(option_dict),
                options.get_schema(self.action),
                'options',
                f'{self.action} singleton action "options"',
            ).result()
            self.options = self.prune_excluded(_)
            # Remove this after the schema check, as the action class won't need
            # it as an arg
            if self.action in ['delete_snapshots', 'restore']:
                del self.options['repository']
        except FailedValidation as exc:
            self.logger.critical('Unable to parse options: %s', exc)
            sys.exit(1)

    def check_filters(self, filter_dict, loc='singleton', key='filters'):
        """Validate provided filters"""
        try:
            self.logger.debug('Validating provided filters: %s', filter_dict)
            _ = SchemaCheck(
                filter_dict,
                Schema(validfilters(self.action, location=loc)),
                key,
                f'{self.action} singleton action "{key}"',
            ).result()
            self.filters = validate_filters(self.action, _)
        except FailedValidation as exc:
            self.logger.critical('Unable to parse filters: %s', exc)
            sys.exit(1)

    def do_filters(self):
        """Actually run the filters"""
        self.logger.debug('Running filters and testing for empty list object')
        if self.allow_ilm:
            self.filters.append({'filtertype': 'ilm', 'exclude': True})
        try:
            self.list_object.iterate_filters({'filters': self.filters})
            self.list_object.empty_list_check()
        except (NoIndices, NoSnapshots) as exc:
            otype = 'index' if isinstance(exc, NoIndices) else 'snapshot'
            if self.ignore:
                self.logger.info('Singleton action not performed: empty %s list', otype)
                sys.exit(0)
            else:
                self.logger.error('Singleton action failed due to empty %s list', otype)
                sys.exit(1)

    def get_list_object(self):
        """Get either a SnapshotList or IndexList object"""
        if self.action in snapshot_actions() or self.action == 'show_snapshots':
            self.list_object = SnapshotList(self.client, repository=self.repository)
        else:
            self.list_object = IndexList(
                self.client, search_pattern=self.search_pattern
            )

    def get_alias_obj(self):
        """Get the Alias object"""
        action_obj = Alias(
            name=self.alias['name'], extra_settings=self.alias['extra_settings']
        )
        for k in ['remove', 'add']:
            if k in self.alias:
                msg = (
                    f"{'Add' if k == 'add' else 'Remov'}ing matching indices "
                    f"{'to' if k == 'add' else 'from'} alias \"{self.alias['name']}\""
                )
                self.logger.debug(msg)
                self.alias[k]['ilo'] = IndexList(
                    self.client, search_pattern=self.search_pattern
                )
                self.alias[k]['ilo'].iterate_filters(
                    {'filters': self.alias[k]['filters']}
                )
                fltr = getattr(action_obj, k)
                fltr(self.alias[k]['ilo'], warn_if_no_indices=self.alias['wini'])
        return action_obj

    def do_singleton_action(self, dry_run=False):
        """Execute the (ostensibly) completely ready to run action"""
        self.logger.debug('Doing the singleton "%s" action here.', self.action)
        try:
            if self.action == 'alias':
                action_obj = self.get_alias_obj()
            elif self.action in ['cluster_routing', 'create_index', 'rollover']:
                action_obj = self.action_class(self.client, **self.options)
            else:
                self.get_list_object()
                self.do_filters()
                self.logger.debug('OPTIONS = %s', self.options)
                action_obj = self.action_class(self.list_object, **self.options)
            if dry_run:
                action_obj.do_dry_run()
            else:
                action_obj.do_action()
        # pylint: disable=broad-except
        except Exception as exc:
            self.logger.critical(
                'Failed to complete action: %s.  %s: %s', self.action, type(exc), exc
            )
            sys.exit(1)
        self.logger.info('"%s" action completed.', self.action)



"""Helper utilities

The kind that don't fit in testers, getters, date_ops, or converters
"""

import re
import logging
from es_client.helpers.utils import ensure_list
from curator.exceptions import FailedExecution

logger = logging.getLogger(__name__)


def chunk_index_list(indices):
    """
    This utility chunks very large index lists into 3KB chunks.
    It measures the size as a csv string, then converts back into a list for the
    return value.

    :param indices: The list of indices

    :type indices: list

    :returns: A list of lists (each a piece of the original ``indices``)
    :rtype: list
    """
    chunks = []
    chunk = ""
    for index in indices:
        if len(chunk) < 3072:
            if not chunk:
                chunk = index
            else:
                chunk += "," + index
        else:
            chunks.append(chunk.split(','))
            chunk = index
    chunks.append(chunk.split(','))
    return chunks


def report_failure(exception):
    """
    Raise a :py:exc:`~.curator.exceptions.FailedExecution` exception and include
    the original error message.

    :param exception: The upstream exception.

    :type exception: :py:exc:Exception

    :rtype: None
    """
    raise FailedExecution(
        f'Exception encountered.  Rerun with loglevel DEBUG and/or check '
        f'Elasticsearch logs for more information. Exception: {exception}'
    )


def show_dry_run(ilo, action, **kwargs):
    """
    Log dry run output with the action which would have been executed.

    :param ilo: An IndexList Object
    :param action: The ``action`` to be performed.
    :param kwargs: Any other args to show in the log output


    :type ilo: :py:class:`~.curator.indexlist.IndexList`
    :type action: str
    :type kwargs: dict

    :rtype: None
    """
    logger.info('DRY-RUN MODE.  No changes will be made.')
    msg = (
        f'(CLOSED) indices may be shown that may not be acted on by action "{action}".'
    )
    logger.info(msg)
    indices = sorted(ilo.indices)
    for idx in indices:
        # Dry runs need index state, so we collect it here if it's not present.
        try:
            index_closed = ilo.index_info[idx]['state'] == 'close'
        except KeyError:
            ilo.get_index_state()
            index_closed = ilo.index_info[idx]['state'] == 'close'
        var = ' (CLOSED)' if index_closed else ''
        msg = f'DRY-RUN: {action}: {idx}{var} with arguments: {kwargs}'
        logger.info(msg)


def to_csv(indices):
    """
    :param indices: A list of indices to act on, or a single value, which could be
        in the format of a csv string already.

    :type indices: list

    :returns: A csv string from a list of indices, or a single value if only one
        value is present
    :rtype: str
    """
    indices = ensure_list(indices)  # in case of a single value passed
    if indices:
        return ','.join(sorted(indices))
    return None


def multitarget_fix(pattern: str) -> str:
    """
    If pattern only has '-' prefixed entries (excludes)
    prepend a wildcard to pattern

    :param pattern: The Elasticsearch multi-target syntax pattern
    :type pattern: str

    :returns: The pattern, possibly with a wildcard prepended
    :rtype: str
    """
    # Split pattern into elements
    elements = pattern.split(',')
    if len(elements) == 1 and elements[0] == '':
        return '*'
    # Initialize positive and negative lists
    positives = []
    negatives = []
    # Loop through elements and sort them into positive and negative lists
    for element in elements:
        if element.startswith('-'):
            negatives.append(element)
        else:
            positives.append(element)
    # If there are no positive elements, but there are negative elements,
    # add a wildcard to the beginning of the pattern
    if len(positives) == 0 and len(negatives) > 0:
        logger.debug(
            "Only negative elements in pattern. Adding '*' so there's something "
            " to remove"
        )
        return '*,' + pattern
    # Otherwise, return the original pattern
    return pattern


def regex_loop(matchstr: str, indices: list) -> list:
    """
    Loop through indices,
    Match against matchstr,
    return matches

    :param matchstr: The Python regex pattern to match against
    :param indices: The list of indices to match against
    :type matchstr: str
    :type indices: list
    :returns: The list of matching indices
    :rtype: list
    """
    retval = []
    for idx in indices:
        if re.match(matchstr, idx):
            retval.append(idx)
    return retval


def multitarget_match(pattern: str, index_list: list) -> list:
    """
    Convert Elasticsearch multi-target syntax ``pattern`` into Python regex
    patterns. Match against ``index_list`` and return the list of matches while
    excluding any negative matches.

    :param pattern: The Elasticsearch multi-target syntax pattern
    :param index_list: The list of indices to match against
    :type pattern: str
    :type index_list: list

    :returns: The final resulting list of indices
    :rtype: list
    """
    retval = []
    excluded = []
    includes = []
    excludes = []
    logger.debug('Multi-target syntax pattern: %s', pattern)
    elements = multitarget_fix(pattern).split(',')
    logger.debug('Individual elements of pattern: %s', elements)
    for element in elements:
        # Exclude elements are prefixed with '-'
        exclude = element.startswith('-')
        # Any index prepended with a . is probably a hidden index, but
        # we need to escape the . for regex to treat it as a literal
        matchstr = element.replace('.', '\\.')
        # Replace Elasticsearch wildcard * with .* for Python regex
        matchstr = matchstr.replace('*', '.*')
        # logger.debug('matchstr: %s', matchstr)
        # If it is not an exclude, add the output of regex_loop to `includes`
        if not exclude:
            includes += regex_loop(matchstr, index_list)
        # If it is an exclude pattern, add the output of regex_loop to `excludes`
        # Remove the `-` from the matchstr ([1:]) before sending.
        if exclude:
            excludes += regex_loop(matchstr[1:], index_list)
    # Create a unique list of indices to loop through
    for idx in list(set(includes)):
        # If idx is not in the unique list of excludes, add it to retval
        if idx not in list(set(excludes)):
            retval.append(idx)
        else:
            # Otherwise, add it to the excludes
            excluded.append(idx)
    # Sort the lists alphabetically
    retval.sort()
    excluded.sort()
    # Log the results
    logger.debug('Included indices: %s', retval)
    logger.debug('Excluded indices: %s', excluded)
    return retval



"""CLI module for curator_cli"""

import click
from es_client.defaults import SHOW_EVERYTHING
from es_client.helpers.config import (
    cli_opts,
    context_settings,
    generate_configdict,
    get_config,
    options_from_dict,
)
from es_client.helpers.logging import configure_logging
from es_client.helpers.utils import option_wrapper
from curator.defaults.settings import CLICK_DRYRUN, default_config_file, footer
from curator._version import __version__
from curator.cli_singletons import (
    alias,
    allocation,
    close,
    delete_indices,
    delete_snapshots,
    forcemerge,
    open_indices,
    replicas,
    restore,
    rollover,
    snapshot,
    shrink,
)
from curator.cli_singletons.show import show_indices, show_snapshots

click_opt_wrap = option_wrapper()


# pylint: disable=R0913, R0914, W0613, W0622, W0718
@click.group(
    context_settings=context_settings(),
    epilog=footer(__version__, tail='singleton-cli.html'),
)
@options_from_dict(SHOW_EVERYTHING)
@click_opt_wrap(*cli_opts('dry-run', settings=CLICK_DRYRUN))
@click.version_option(__version__, '-v', '--version', prog_name='curator_cli')
@click.pass_context
def curator_cli(
    ctx,
    config,
    hosts,
    cloud_id,
    api_token,
    id,
    api_key,
    username,
    password,
    bearer_auth,
    opaque_id,
    request_timeout,
    http_compress,
    verify_certs,
    ca_certs,
    client_cert,
    client_key,
    ssl_assert_hostname,
    ssl_assert_fingerprint,
    ssl_version,
    master_only,
    skip_version_test,
    loglevel,
    logfile,
    logformat,
    blacklist,
    dry_run,
):
    """
    Curator CLI (Singleton Tool)

    Run a single action from the command-line.

    The default $HOME/.curator/curator.yml configuration file (--config)
    can be used but is not needed.

    Command-line settings will always override YAML configuration settings.
    """
    ctx.obj = {}
    ctx.obj['dry_run'] = dry_run
    ctx.obj['default_config'] = default_config_file()
    get_config(ctx)
    configure_logging(ctx)
    generate_configdict(ctx)


# Add the subcommands
curator_cli.add_command(alias)
curator_cli.add_command(allocation)
curator_cli.add_command(close)
curator_cli.add_command(delete_indices)
curator_cli.add_command(delete_snapshots)
curator_cli.add_command(forcemerge)
curator_cli.add_command(open_indices)
curator_cli.add_command(replicas)
curator_cli.add_command(snapshot)
curator_cli.add_command(restore)
curator_cli.add_command(rollover)
curator_cli.add_command(shrink)
curator_cli.add_command(show_indices)
curator_cli.add_command(show_snapshots)



"""Tending your Elasticsearch indices and snapshots"""
from curator._version import __version__
from curator.helpers import *
from curator.exceptions import *
from curator.defaults import *
from curator.validators import *
from curator.indexlist import IndexList
from curator.snapshotlist import SnapshotList
from curator.actions import *
from curator.cli import *
from curator.repomgrcli import *



"""Other Classes"""

import logging
from es_client.exceptions import FailedValidation
from es_client.helpers.schemacheck import password_filter
from es_client.helpers.utils import get_yaml
from curator import IndexList, SnapshotList
from curator.actions import CLASS_MAP
from curator.exceptions import ConfigurationError
from curator.helpers.testers import validate_actions

# Let me tell you the story of the nearly wasted afternoon and the research that went
# into this seemingly simple work-around. Actually, no. It's even more wasted time
# writing that story here. Suffice to say that I couldn't use the CLASS_MAP with class
# objects to directly map them to class instances. The Wrapper class and the
# ActionDef.instantiate method do all of the work for me, allowing me to easily and
# cleanly pass *args and **kwargs to the individual action classes of CLASS_MAP.


class Wrapper:
    """Wrapper Class"""

    def __init__(self, cls):
        """Instantiate with passed Class (not instance or object)

        :param cls: A class (not an instance of the class)
        """
        #: The class itself (not an instance of it), passed from ``cls``
        self.class_object = cls
        #: An instance of :py:attr:`class_object`
        self.class_instance = None

    def set_instance(self, *args, **kwargs):
        """Set up :py:attr:`class_instance` from :py:attr:`class_object`"""
        self.class_instance = self.class_object(*args, **kwargs)

    def get_instance(self, *args, **kwargs):
        """Return the instance with ``*args`` and ``**kwargs``"""
        self.set_instance(*args, **kwargs)
        return self.class_instance


class ActionsFile:
    """Class to parse and verify entire actions file

    Individual actions are :py:class:`~.curator.classdef.ActionDef` objects
    """

    def __init__(self, action_file):
        self.logger = logging.getLogger(__name__)
        #: The full, validated configuration from ``action_file``.
        self.fullconfig = self.get_validated(action_file)
        self.logger.debug('Action Configuration: %s', password_filter(self.fullconfig))
        #: A dict of all actions in the provided configuration. Each original key name
        #: is preserved and the value is now an
        #: :py:class:`~.curator.classdef.ActionDef`, rather than a dict.
        self.actions = None
        self.set_actions(self.fullconfig['actions'])

    def get_validated(self, action_file):
        """
        :param action_file: The path to a valid YAML action configuration file
        :type action_file: str

        :returns: The result from passing ``action_file`` to
            :py:func:`~.curator.helpers.testers.validate_actions`
        """
        try:
            return validate_actions(get_yaml(action_file))
        except (FailedValidation, UnboundLocalError) as err:
            self.logger.critical('Configuration Error: %s', err)
            raise ConfigurationError from err

    def parse_actions(self, all_actions):
        """Parse the individual actions found in ``all_actions['actions']``

        :param all_actions: All actions, each its own dictionary behind a numeric key.
            Making the keys numeric guarantees that if they are sorted, they will
            always be executed in order.

        :type all_actions: dict

        :returns:
        :rtype: list of :py:class:`~.curator.classdef.ActionDef`
        """
        acts = {}
        for idx in all_actions.keys():
            acts[idx] = ActionDef(all_actions[idx])
        return acts

    def set_actions(self, all_actions):
        """Set the actions via :py:meth:`~.curator.classdef.ActionsFile.parse_actions`

        :param all_actions: All actions, each its own dictionary behind a numeric key.
            Making the keys numeric guarantees that if they are sorted, they will
            always be executed in order.
        :type all_actions: dict
        :rtype: None
        """
        self.actions = self.parse_actions(all_actions)


# In this case, I just don't care that pylint thinks I'm overdoing it with attributes
# pylint: disable=too-many-instance-attributes
class ActionDef:
    """Individual Action Definition Class

    Instances of this class represent an individual action from an action file.
    """

    def __init__(self, action_dict):
        #: The whole action dictionary
        self.action_dict = action_dict
        #: The action name
        self.action = None
        #: The action's class (Alias, Allocation, etc.)
        self.action_cls = None
        #: Only when action is alias will this be a :py:class:`~.curator.IndexList`
        self.alias_adds = None
        #: Only when action is alias will this be a :py:class:`~.curator.IndexList`
        self.alias_removes = None
        #: The list class, either :py:class:`~.curator.IndexList` or
        #: :py:class:`~.curator.SnapshotList`. Default is
        #: :py:class:`~.curator.IndexList`
        self.list_obj = Wrapper(IndexList)
        #: The action ``description``
        self.description = None
        #: The action ``options`` :py:class:`dict`
        self.options = {}
        #: The action ``filters`` :py:class:`list`
        self.filters = None
        #: The action option ``disable_action``
        self.disabled = None
        #: The action option ``continue_if_exception``
        self.cif = None
        #: The action option ``timeout_override``
        self.timeout_override = None
        #: The action option ``ignore_empty_list``
        self.iel = None
        #: The action option ``allow_ilm_indices``
        self.allow_ilm = None
        self.set_root_attrs()
        self.set_option_attrs()
        self.log_the_options()
        self.get_action_class()

    def instantiate(self, attribute, *args, **kwargs):
        """
        Convert ``attribute`` from being a :py:class:`~.curator.classdef.Wrapper` of a
        Class to an instantiated object of that Class.

        This is madness or genius. You decide. This entire method plus the
        :py:class:`~.curator.classdef.Wrapper` class came about because I couldn't
        cleanly instantiate a class variable into a class object. It works, and that's
        good enough for me.

        :param attribute: The `name` of an attribute that references a Wrapper class
            instance
        :type attribute: str
        """
        try:
            wrapper = getattr(self, attribute)
        except AttributeError as exc:
            raise AttributeError(
                f'Bad Attribute: {attribute}. Exception: {exc}'
            ) from exc
        setattr(self, attribute, self.get_obj_instance(wrapper, *args, **kwargs))

    def get_obj_instance(self, wrapper, *args, **kwargs):
        """Get the class instance wrapper identified by ``wrapper``
        Pass all other args and kwargs to the
        :py:meth:`~.curator.classdef.Wrapper.get_instance` method.

        :returns: An instance of the class that :py:class:`~.curator.classdef.Wrapper`
            is wrapping
        """
        if not isinstance(wrapper, Wrapper):
            raise ConfigurationError(
                f'{__name__} was passed wrapper which was of type {type(wrapper)}'
            )
        return wrapper.get_instance(*args, **kwargs)

    def set_alias_extras(self):
        """Populate the :py:attr:`alias_adds` and :py:attr:`alias_removes` attributes"""
        self.alias_adds = Wrapper(IndexList)
        self.alias_removes = Wrapper(IndexList)

    def get_action_class(self):
        """Get the action class from :py:const:`~.curator.actions.CLASS_MAP`

        Do extra setup when action is ``alias``

        Set :py:attr:`list_obj` to :py:class:`~.curator.SnapshotList` when
        :py:attr:`~.curator.classdef.ActionDef.action` is ``delete_snapshots`` or
        ``restore``
        """

        self.action_cls = Wrapper(CLASS_MAP[self.action])
        if self.action == 'alias':
            self.set_alias_extras()
        if self.action in ['delete_snapshots', 'restore']:
            self.list_obj = Wrapper(SnapshotList)

    def set_option_attrs(self):
        """
        Iteratively get the keys and values from
        :py:attr:`~.curator.classdef.ActionDef.options` and set the attributes
        """
        attmap = {
            'disable_action': 'disabled',
            'continue_if_exception': 'cif',
            'ignore_empty_list': 'iel',
            'allow_ilm_indices': 'allow_ilm',
            'timeout_override': 'timeout_override',
        }
        for key in self.action_dict['options']:
            if key in attmap:
                setattr(self, attmap[key], self.action_dict['options'][key])
            else:
                self.options[key] = self.action_dict['options'][key]

    def set_root_attrs(self):
        """
        Iteratively get the keys and values from
        :py:attr:`~.curator.classdef.ActionDef.action_dict` and set the attributes
        """
        for key, value in self.action_dict.items():
            # Gonna grab options in get_option_attrs()
            if key == 'options':
                continue
            if value is not None:
                setattr(self, key, value)

    def log_the_options(self):
        """Log options at initialization time"""
        logger = logging.getLogger('curator.cli.ActionDef')
        msg = (
            f'For action {self.action}: disable_action={self.disabled}'
            f'continue_if_exception={self.cif}, '
            f'timeout_override={self.timeout_override}, '
            f'ignore_empty_list={self.iel}, allow_ilm_indices={self.allow_ilm}'
        )
        logger.debug(msg)
        if self.allow_ilm:
            logger.warning('Permitting operation on indices with an ILM policy')



"""Create index action class"""

import logging

# pylint: disable=import-error, broad-except
from elasticsearch8.exceptions import RequestError
from curator.exceptions import ConfigurationError, FailedExecution
from curator.helpers.date_ops import parse_date_pattern
from curator.helpers.utils import report_failure


class CreateIndex:
    """Create Index Action Class"""

    def __init__(self, client, name=None, extra_settings=None, ignore_existing=False):
        """
        :param client: A client connection object
        :param name: A name, which can contain :py:func:`time.strftime` strings
        :param extra_settings: The `settings` and `mappings` for the index. For
            more information see `the create indices documentation
            </https://www.elastic.co/guide/en/elasticsearch/reference/8.6/indices-create-index.html>`_.
        :param ignore_existing: If an index already exists, and this setting is
            ``True``, ignore the 400 error that results in a
            ``resource_already_exists_exception`` and return that it was successful.

        :type client: :py:class:`~.elasticsearch.Elasticsearch`
        :type name: str
        :type extra_settings: dict
        :type ignore_existing: bool
        """
        if extra_settings is None:
            extra_settings = {}
        if not name:
            raise ConfigurationError('Value for "name" not provided.')
        #: The :py:func:`~.curator.helpers.date_ops.parse_date_pattern` rendered
        #: version of what was passed as ``name``.
        self.name = parse_date_pattern(name)
        #: Extracted from the action definition, it should be a boolean informing
        #: whether to ignore the error if the index already exists.
        self.ignore_existing = ignore_existing
        #: An :py:class:`~.elasticsearch.Elasticsearch` client object
        self.client = client
        #: Any extra settings for the index, like aliases, mappings, or settings.
        #: Gets the value from param ``extra_settings``.
        self.extra_settings = extra_settings
        #: Gets any ``aliases`` from :py:attr:`extra_settings` or is :py:class:`None`
        self.aliases = None
        #: Gets any ``mappings`` from :py:attr:`extra_settings` or is :py:class:`None`
        self.mappings = None
        #: Gets any ``settings`` from :py:attr:`extra_settings` or is :py:class:`None`
        self.settings = None

        if 'aliases' in extra_settings:
            self.aliases = extra_settings.pop('aliases')
        if 'mappings' in extra_settings:
            self.mappings = extra_settings.pop('mappings')
        if 'settings' in extra_settings:
            self.settings = extra_settings.pop('settings')
        self.loggit = logging.getLogger('curator.actions.create_index')

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        msg = (
            f'DRY-RUN: create_index "{self.name}" with arguments: {self.extra_settings}'
        )
        self.loggit.info(msg)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.create` index identified
        by :py:attr:`name` with values from :py:attr:`aliases`, :py:attr:`mappings`,
        and :py:attr:`settings`
        """
        msg = f'Creating index "{self.name}" with settings: {self.extra_settings}'
        self.loggit.info(msg)
        try:
            self.client.indices.create(
                index=self.name,
                aliases=self.aliases,
                mappings=self.mappings,
                settings=self.settings,
            )
        # Most likely error is a 400, `resource_already_exists_exception`
        except RequestError as err:
            match_list = [
                "index_already_exists_exception",
                "resource_already_exists_exception",
            ]
            if err.error in match_list:
                if self.ignore_existing:
                    self.loggit.warning('Index %s already exists.', self.name)
                else:
                    raise FailedExecution(f'Index {self.name} already exists.') from err
            else:
                msg = f'Unable to create index "{self.name}". Error: {err.error}'
                raise FailedExecution(msg) from err
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)
